# Copyright 2016 Red Hat | Ansible
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import os
import platform
import re
import sys
import traceback
import typing as t
from collections.abc import Mapping, Sequence

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils.parsing.convert_bool import BOOLEANS_FALSE, BOOLEANS_TRUE

from ansible_collections.community.docker.plugins.module_utils._util import (
    DEFAULT_DOCKER_HOST,
    DEFAULT_TIMEOUT_SECONDS,
    DEFAULT_TLS,
    DEFAULT_TLS_VERIFY,
    DOCKER_COMMON_ARGS,
    DOCKER_MUTUALLY_EXCLUSIVE,
    DOCKER_REQUIRED_TOGETHER,
    sanitize_result,
    update_tls_hostname,
)
from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)

HAS_DOCKER_PY_2 = False  # pylint: disable=invalid-name
HAS_DOCKER_PY_3 = False  # pylint: disable=invalid-name
HAS_DOCKER_ERROR: None | str  # pylint: disable=invalid-name
HAS_DOCKER_TRACEBACK: None | str  # pylint: disable=invalid-name
docker_version: str | None  # pylint: disable=invalid-name

try:
    from docker import __version__ as docker_version
    from docker.errors import APIError, TLSParameterError
    from docker.tls import TLSConfig

    if LooseVersion(docker_version) >= LooseVersion("3.0.0"):
        HAS_DOCKER_PY_3 = True  # pylint: disable=invalid-name
        from docker import APIClient as Client
    elif LooseVersion(docker_version) >= LooseVersion("2.0.0"):
        HAS_DOCKER_PY_2 = True  # pylint: disable=invalid-name
        from docker import APIClient as Client
    else:
        from docker import Client  # type: ignore

except ImportError as exc:
    HAS_DOCKER_ERROR = str(exc)  # pylint: disable=invalid-name
    HAS_DOCKER_TRACEBACK = traceback.format_exc()  # pylint: disable=invalid-name
    HAS_DOCKER_PY = False  # pylint: disable=invalid-name
    docker_version = None  # pylint: disable=invalid-name
else:
    HAS_DOCKER_PY = True  # pylint: disable=invalid-name
    HAS_DOCKER_ERROR = None  # pylint: disable=invalid-name
    HAS_DOCKER_TRACEBACK = None  # pylint: disable=invalid-name


try:
    from requests.exceptions import (  # noqa: F401, pylint: disable=unused-import
        RequestException,
    )
except ImportError:
    # Either Docker SDK for Python is no longer using requests, or Docker SDK for Python is not around either,
    # or Docker SDK for Python's dependency requests is missing. In any case, define an exception
    # class RequestException so that our code does not break.
    class RequestException(Exception):  # type: ignore
        pass


if t.TYPE_CHECKING:
    from collections.abc import Callable


MIN_DOCKER_VERSION = "2.0.0"


if not HAS_DOCKER_PY:
    # No Docker SDK for Python. Create a place holder client to allow
    # instantiation of AnsibleModule and proper error handing
    class Client:  # type: ignore # noqa: F811, pylint: disable=function-redefined
        def __init__(self, **kwargs: t.Any) -> None:
            pass

    class APIError(Exception):  # type: ignore # noqa: F811, pylint: disable=function-redefined
        pass

    class NotFound(Exception):  # type: ignore # noqa: F811, pylint: disable=function-redefined
        pass


def _get_tls_config(
    fail_function: Callable[[str], t.NoReturn], **kwargs: t.Any
) -> TLSConfig:
    if "assert_hostname" in kwargs and LooseVersion(docker_version) >= LooseVersion(
        "7.0.0b1"
    ):
        assert_hostname = kwargs.pop("assert_hostname")
        if assert_hostname is not None:
            fail_function(
                "tls_hostname is not compatible with Docker SDK for Python 7.0.0+. You are using"
                f" Docker SDK for Python {docker_version}. The tls_hostname option (value: {assert_hostname})"
                " has either been set directly or with the environment variable DOCKER_TLS_HOSTNAME."
                " Make sure it is not set, or switch to an older version of Docker SDK for Python."
            )
    # Filter out all None parameters
    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    try:
        return TLSConfig(**kwargs)
    except TLSParameterError as exc:
        fail_function(f"TLS config error: {exc}")


def is_using_tls(auth_data: dict[str, t.Any]) -> bool:
    return auth_data["tls_verify"] or auth_data["tls"]


def get_connect_params(
    auth_data: dict[str, t.Any], fail_function: Callable[[str], t.NoReturn]
) -> dict[str, t.Any]:
    if is_using_tls(auth_data):
        auth_data["docker_host"] = auth_data["docker_host"].replace(
            "tcp://", "https://"
        )

    result = {
        "base_url": auth_data["docker_host"],
        "version": auth_data["api_version"],
        "timeout": auth_data["timeout"],
    }

    if auth_data["tls_verify"]:
        # TLS with verification
        tls_config: dict[str, t.Any] = {
            "verify": True,
        }
        if auth_data["tls_hostname"] is not None:
            tls_config["assert_hostname"] = auth_data["tls_hostname"]
        if auth_data["cert_path"] and auth_data["key_path"]:
            tls_config["client_cert"] = (auth_data["cert_path"], auth_data["key_path"])
        if auth_data["cacert_path"]:
            tls_config["ca_cert"] = auth_data["cacert_path"]
        result["tls"] = _get_tls_config(fail_function=fail_function, **tls_config)
    elif auth_data["tls"]:
        # TLS without verification
        tls_config = {
            "verify": False,
        }
        if auth_data["cert_path"] and auth_data["key_path"]:
            tls_config["client_cert"] = (auth_data["cert_path"], auth_data["key_path"])
        result["tls"] = _get_tls_config(fail_function=fail_function, **tls_config)

    if auth_data.get("use_ssh_client"):
        if LooseVersion(docker_version) < LooseVersion("4.4.0"):
            fail_function(
                "use_ssh_client=True requires Docker SDK for Python 4.4.0 or newer"
            )
        result["use_ssh_client"] = True

    # No TLS
    return result


DOCKERPYUPGRADE_SWITCH_TO_DOCKER = (
    "Try `pip uninstall docker-py` followed by `pip install docker`."
)
DOCKERPYUPGRADE_UPGRADE_DOCKER = "Use `pip install --upgrade docker` to upgrade."


class AnsibleDockerClientBase(Client):
    def __init__(
        self,
        min_docker_version: str | None = None,
        min_docker_api_version: str | None = None,
    ) -> None:
        if min_docker_version is None:
            min_docker_version = MIN_DOCKER_VERSION

        self.docker_py_version = LooseVersion(docker_version)

        if not HAS_DOCKER_PY:
            msg = missing_required_lib("Docker SDK for Python: docker>=5.0.0")
            msg = f"{msg}, for example via `pip install docker`. The error was: {HAS_DOCKER_ERROR}"
            self.fail(msg, exception=HAS_DOCKER_TRACEBACK)

        if self.docker_py_version < LooseVersion(min_docker_version):
            msg = (
                f"Error: Docker SDK for Python version is {docker_version} ({platform.node()}'s Python {sys.executable})."
                f" Minimum version required is {min_docker_version}."
            )
            if docker_version < LooseVersion("2.0"):
                msg += DOCKERPYUPGRADE_SWITCH_TO_DOCKER
            else:
                msg += DOCKERPYUPGRADE_UPGRADE_DOCKER
            self.fail(msg)

        self._connect_params = get_connect_params(
            self.auth_params, fail_function=self.fail
        )

        try:
            super().__init__(**self._connect_params)
            self.docker_api_version_str = self.api_version
        except APIError as exc:
            self.fail(f"Docker API error: {exc}")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(f"Error connecting: {exc}")

        self.docker_api_version = LooseVersion(self.docker_api_version_str)
        min_docker_api_version = min_docker_api_version or "1.25"
        if self.docker_api_version < LooseVersion(min_docker_api_version):
            self.fail(
                f"Docker API version is {self.docker_api_version_str}. Minimum version required is {min_docker_api_version}."
            )

    def log(self, msg: t.Any, pretty_print: bool = False) -> None:
        pass
        # if self.debug:
        #     from .util import log_debug
        #     log_debug(msg, pretty_print=pretty_print)

    @abc.abstractmethod
    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        pass

    @abc.abstractmethod
    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        pass

    @staticmethod
    def _get_value(
        param_name: str,
        param_value: t.Any,
        env_variable: str | None,
        default_value: t.Any | None,
        value_type: t.Literal["str", "bool", "int"] = "str",
    ) -> t.Any:
        if param_value is not None:
            # take module parameter value
            if value_type == "bool":
                if param_value in BOOLEANS_TRUE:
                    return True
                if param_value in BOOLEANS_FALSE:
                    return False
                return bool(param_value)
            if value_type == "int":
                return int(param_value)
            return param_value

        if env_variable is not None:
            env_value = os.environ.get(env_variable)
            if env_value is not None:
                # take the env variable value
                if param_name == "cert_path":
                    return os.path.join(env_value, "cert.pem")
                if param_name == "cacert_path":
                    return os.path.join(env_value, "ca.pem")
                if param_name == "key_path":
                    return os.path.join(env_value, "key.pem")
                if value_type == "bool":
                    if env_value in BOOLEANS_TRUE:
                        return True
                    if env_value in BOOLEANS_FALSE:
                        return False
                    return bool(env_value)
                if value_type == "int":
                    return int(env_value)
                return env_value

        # take the default
        return default_value

    @abc.abstractmethod
    def _get_params(self) -> dict[str, t.Any]:
        pass

    @property
    def auth_params(self) -> dict[str, t.Any]:
        # Get authentication credentials.
        # Precedence: module parameters-> environment variables-> defaults.

        self.log("Getting credentials")

        client_params = self._get_params()

        params = {}
        for key in DOCKER_COMMON_ARGS:
            params[key] = client_params.get(key)

        result = {
            "docker_host": self._get_value(
                "docker_host",
                params["docker_host"],
                "DOCKER_HOST",
                DEFAULT_DOCKER_HOST,
                value_type="str",
            ),
            "tls_hostname": self._get_value(
                "tls_hostname",
                params["tls_hostname"],
                "DOCKER_TLS_HOSTNAME",
                None,
                value_type="str",
            ),
            "api_version": self._get_value(
                "api_version",
                params["api_version"],
                "DOCKER_API_VERSION",
                "auto",
                value_type="str",
            ),
            "cacert_path": self._get_value(
                "cacert_path",
                params["ca_path"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "cert_path": self._get_value(
                "cert_path",
                params["client_cert"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "key_path": self._get_value(
                "key_path",
                params["client_key"],
                "DOCKER_CERT_PATH",
                None,
                value_type="str",
            ),
            "tls": self._get_value(
                "tls", params["tls"], "DOCKER_TLS", DEFAULT_TLS, value_type="bool"
            ),
            "tls_verify": self._get_value(
                "validate_certs",
                params["validate_certs"],
                "DOCKER_TLS_VERIFY",
                DEFAULT_TLS_VERIFY,
                value_type="bool",
            ),
            "timeout": self._get_value(
                "timeout",
                params["timeout"],
                "DOCKER_TIMEOUT",
                DEFAULT_TIMEOUT_SECONDS,
                value_type="int",
            ),
            "use_ssh_client": self._get_value(
                "use_ssh_client",
                params["use_ssh_client"],
                None,
                False,
                value_type="bool",
            ),
        }

        if LooseVersion(docker_version) < LooseVersion("7.0.0b1"):
            update_tls_hostname(result)

        return result

    def _handle_ssl_error(self, error: Exception) -> t.NoReturn:
        match = re.match(r"hostname.*doesn\'t match (\'.*\')", str(error))
        if match:
            hostname = self.auth_params["tls_hostname"]
            self.fail(
                f"You asked for verification that Docker daemons certificate's hostname matches {hostname}. "
                f"The actual certificate's hostname is {match.group(1)}. Most likely you need to set DOCKER_TLS_HOSTNAME "
                f"or pass `tls_hostname` with a value of {match.group(1)}. You may also use TLS without verification by "
                "setting the `tls` parameter to true."
            )
        self.fail(f"SSL Exception: {error}")


class AnsibleDockerClient(AnsibleDockerClientBase):
    def __init__(
        self,
        argument_spec: dict[str, t.Any] | None = None,
        supports_check_mode: bool = False,
        mutually_exclusive: Sequence[Sequence[str]] | None = None,
        required_together: Sequence[Sequence[str]] | None = None,
        required_if: (
            Sequence[
                tuple[str, t.Any, Sequence[str]]
                | tuple[str, t.Any, Sequence[str], bool]
            ]
            | None
        ) = None,
        required_one_of: Sequence[Sequence[str]] | None = None,
        required_by: dict[str, Sequence[str]] | None = None,
        min_docker_version: str | None = None,
        min_docker_api_version: str | None = None,
        option_minimal_versions: dict[str, t.Any] | None = None,
        option_minimal_versions_ignore_params: Sequence[str] | None = None,
        fail_results: dict[str, t.Any] | None = None,
    ):
        # Modules can put information in here which will always be returned
        # in case client.fail() is called.
        self.fail_results = fail_results or {}

        merged_arg_spec = {}
        merged_arg_spec.update(DOCKER_COMMON_ARGS)
        if argument_spec:
            merged_arg_spec.update(argument_spec)
            self.arg_spec = merged_arg_spec

        mutually_exclusive_params: list[Sequence[str]] = []
        mutually_exclusive_params += DOCKER_MUTUALLY_EXCLUSIVE
        if mutually_exclusive:
            mutually_exclusive_params += mutually_exclusive

        required_together_params: list[Sequence[str]] = []
        required_together_params += DOCKER_REQUIRED_TOGETHER
        if required_together:
            required_together_params += required_together

        self.module = AnsibleModule(
            argument_spec=merged_arg_spec,
            supports_check_mode=supports_check_mode,
            mutually_exclusive=mutually_exclusive_params,
            required_together=required_together_params,
            required_if=required_if,
            required_one_of=required_one_of,
            required_by=required_by or {},
        )

        self.debug = self.module.params.get("debug")
        self.check_mode = self.module.check_mode

        super().__init__(
            min_docker_version=min_docker_version,
            min_docker_api_version=min_docker_api_version,
        )

        if option_minimal_versions is not None:
            self._get_minimal_versions(
                option_minimal_versions, option_minimal_versions_ignore_params
            )

    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        self.fail_results.update(kwargs)
        self.module.fail_json(msg=msg, **sanitize_result(self.fail_results))

    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
    ) -> None:
        self.module.deprecate(
            msg, version=version, date=date, collection_name=collection_name
        )

    def _get_params(self) -> dict[str, t.Any]:
        return self.module.params

    def _get_minimal_versions(
        self,
        option_minimal_versions: dict[str, t.Any],
        ignore_params: Sequence[str] | None = None,
    ) -> None:
        self.option_minimal_versions: dict[str, dict[str, t.Any]] = {}
        for option in self.module.argument_spec:
            if ignore_params is not None and option in ignore_params:
                continue
            self.option_minimal_versions[option] = {}
        self.option_minimal_versions.update(option_minimal_versions)

        for option, data in self.option_minimal_versions.items():
            # Test whether option is supported, and store result
            support_docker_py = True
            support_docker_api = True
            if "docker_py_version" in data:
                support_docker_py = self.docker_py_version >= LooseVersion(
                    data["docker_py_version"]
                )
            if "docker_api_version" in data:
                support_docker_api = self.docker_api_version >= LooseVersion(
                    data["docker_api_version"]
                )
            data["supported"] = support_docker_py and support_docker_api
            # Fail if option is not supported but used
            if not data["supported"]:
                # Test whether option is specified
                if "detect_usage" in data:
                    used = data["detect_usage"](self)
                else:
                    used = self.module.params.get(option) is not None
                    if used and "default" in self.module.argument_spec[option]:
                        used = (
                            self.module.params[option]
                            != self.module.argument_spec[option]["default"]
                        )
                if used:
                    # If the option is used, compose error message.
                    if "usage_msg" in data:
                        usg = data["usage_msg"]
                    else:
                        usg = f"set {option} option"
                    if not support_docker_api:
                        msg = f"Docker API version is {self.docker_api_version_str}. Minimum version required is {data['docker_api_version']} to {usg}."
                    elif not support_docker_py:
                        msg = (
                            f"Docker SDK for Python version is {docker_version} ({platform.node()}'s Python {sys.executable})."
                            f" Minimum version required is {data['docker_py_version']} to {usg}. {DOCKERPYUPGRADE_UPGRADE_DOCKER}"
                        )
                    else:
                        # should not happen
                        msg = f"Cannot {usg} with your configuration."
                    self.fail(msg)

    def report_warnings(
        self, result: t.Any, warnings_key: Sequence[str] | None = None
    ) -> None:
        """
        Checks result of client operation for warnings, and if present, outputs them.

        warnings_key should be a list of keys used to crawl the result dictionary.
        For example, if warnings_key == ['a', 'b'], the function will consider
        result['a']['b'] if these keys exist. If the result is a non-empty string, it
        will be reported as a warning. If the result is a list, every entry will be
        reported as a warning.

        In most cases (if warnings are returned at all), warnings_key should be
        ['Warnings'] or ['Warning']. The default value (if not specified) is ['Warnings'].
        """
        if warnings_key is None:
            warnings_key = ["Warnings"]
        for key in warnings_key:
            if not isinstance(result, Mapping):
                return
            result = result.get(key)
        if isinstance(result, Sequence):
            for warning in result:
                self.module.warn(f"Docker warning: {warning}")
        elif isinstance(result, str) and result:
            self.module.warn(f"Docker warning: {result}")



# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Jonathan Mainguy <jon@soh.re>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type


class ModuleDocFragment(object):

    # Standard mysql documentation fragment
    DOCUMENTATION = r'''
options:
  login_user:
    description:
      - The username used to authenticate with.
    type: str
  login_password:
    description:
      - The password used to authenticate with.
    type: str
  login_host:
    description:
      - Host running the database.
      - In some cases for local connections the I(login_unix_socket=/path/to/mysqld/socket),
        that is usually C(/var/run/mysqld/mysqld.sock), needs to be used instead of I(login_host=localhost).
    type: str
    default: localhost
  login_port:
    description:
      - Port of the MySQL server. Requires I(login_host) be defined as other than localhost if login_port is used.
    type: int
    default: 3306
  login_unix_socket:
    description:
      - The path to a Unix domain socket for local connections.
      - Use this parameter to avoid the C(Please explicitly state intended protocol) error.
    type: str
  connect_timeout:
    description:
      - The connection timeout when connecting to the MySQL server.
    type: int
    default: 30
  config_file:
    description:
      - Specify a config file from which user and password are to be read.
      - The default config file, C(~/.my.cnf), if it exists, will be read, even if I(config_file) is not specified.
      - The default config file, C(~/.my.cnf), if it exists, must contain a C([client]) section as a MySQL connector requirement.
      - To prevent the default config file from being read, set I(config_file) to be an empty string.
    type: path
    default: '~/.my.cnf'
  ca_cert:
    description:
      - The path to a Certificate Authority (CA) certificate. This option, if used, must specify the same certificate
        as used by the server.
    type: path
    aliases: [ ssl_ca ]
  client_cert:
    description:
      - The path to a client public key certificate.
    type: path
    aliases: [ ssl_cert ]
  client_key:
    description:
      - The path to the client private key.
    type: path
    aliases: [ ssl_key ]
  check_hostname:
    description:
      - Whether to validate the server host name when an SSL connection is required. Corresponds to MySQL CLIs C(--ssl) switch.
      - Setting this to C(false) disables hostname verification. Use with caution.
      - Requires pymysql >= 0.7.11.
    type: bool
    version_added: '1.1.0'
requirements:
   - PyMySQL (Python 2.7 and Python 3.x)
notes:
   - Requires the PyMySQL (Python 2.7 and Python 3.X) package installed on the remote host.
     The Python package may be installed with apt-get install python-pymysql (Ubuntu; see M(ansible.builtin.apt)) or
     yum install python2-PyMySQL (RHEL/CentOS/Fedora; see M(ansible.builtin.yum)). You can also use dnf install python2-PyMySQL
     for newer versions of Fedora; see M(ansible.builtin.dnf).
   - Be sure you have PyMySQL library installed on the target machine
     for the Python interpreter Ansible discovers. For example if ansible discovers and uses Python 3, you need to install
     the Python 3 version of PyMySQL. If ansible discovers and uses Python 2, you need to install the Python 2
     version of PyMySQL.
   - If you have trouble, it may help to force Ansible to use the Python interpreter you need by specifying
     C(ansible_python_interpreter). For more information, see
     U(https://docs.ansible.com/ansible/latest/reference_appendices/interpreter_discovery.html).
   - Both C(login_password) and C(login_user) are required when you are
     passing credentials. If none are present, the module will attempt to read
     the credentials from C(~/.my.cnf), and finally fall back to using the MySQL
     default login of 'root' with no password.
   - If there are problems with local connections, using I(login_unix_socket=/path/to/mysqld/socket)
     instead of I(login_host=localhost) might help. As an example, the default MariaDB installation of version 10.4
     and later uses the unix_socket authentication plugin by default that
     without using I(login_unix_socket=/var/run/mysqld/mysqld.sock) (the default path)
     causes the error ``Host '127.0.0.1' is not allowed to connect to this MariaDB server``.
   - "If credentials from the config file (for example, C(/root/.my.cnf)) are not needed to connect to a database server, but
     the file exists and does not contain a C([client]) section, before any other valid directives, it will be read and this
     will cause the connection to fail, to prevent this set it to an empty string, (for example C(config_file: ''))."
   - "To avoid the C(Please explicitly state intended protocol) error, use the I(login_unix_socket) argument,
     for example, C(login_unix_socket: /run/mysqld/mysqld.sock)."
   - Alternatively, to avoid using I(login_unix_socket) argument on each invocation you can specify the socket path
     using the `socket` option in your MySQL config file (usually C(~/.my.cnf)) on the destination host, for
     example C(socket=/var/lib/mysql/mysql.sock).
attributes:
  check_mode:
    description: Can run in check_mode and return changed status prediction without modifying target.
  idempotent:
    description:
      - When run twice in a row outside check mode, with the same arguments, the second invocation indicates no change.
      - This assumes that the system controlled/queried by the module has not changed in a relevant way.
'''



# -*- coding: utf-8 -*-

#  Copyright 2017 Michael De La Rue | Ansible
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import re


def validate_aws_arn(
    arn, partition=None, service=None, region=None, account_id=None, resource=None, resource_type=None, resource_id=None
):
    details = parse_aws_arn(arn)

    if not details:
        return False

    if partition and details.get("partition") != partition:
        return False
    if service and details.get("service") != service:
        return False
    if region and details.get("region") != region:
        return False
    if account_id and details.get("account_id") != account_id:
        return False
    if resource and details.get("resource") != resource:
        return False
    if resource_type and details.get("resource_type") != resource_type:
        return False
    if resource_id and details.get("resource_id") != resource_id:
        return False

    return True


def parse_aws_arn(arn):
    """
    Based on https://docs.aws.amazon.com/IAM/latest/UserGuide/reference-arns.html

    The following are the general formats for ARNs.
        arn:partition:service:region:account-id:resource-id
        arn:partition:service:region:account-id:resource-type/resource-id
        arn:partition:service:region:account-id:resource-type:resource-id
    The specific formats depend on the resource.
    The ARNs for some resources omit the Region, the account ID, or both the Region and the account ID.

    Note: resource_type handling is very naive, for complex cases it may be necessary to use
    "resource" directly instead of resource_type, this will include the resource type and full ID,
    including all paths.
    """
    m = re.search(r"arn:(aws(-([a-z\-]+))?):([\w-]+):([a-z0-9\-]*):(\d*|aws|aws-managed):(.*)", arn)
    if m is None:
        return None
    result = dict()
    result.update(dict(partition=m.group(1)))
    result.update(dict(service=m.group(4)))
    result.update(dict(region=m.group(5)))
    result.update(dict(account_id=m.group(6)))
    result.update(dict(resource=m.group(7)))

    m2 = re.search(r"^([^:/]*)[:/](.+)$", m.group(7))
    if m2 is None:
        result.update(dict(resource_type=None, resource_id=m.group(7)))
    else:
        result.update(dict(resource_type=m2.group(1), resource_id=m2.group(2)))

    return result


# An implementation of this used was originally in ec2.py, however Outposts
# aren't specific to the EC2 service
def is_outpost_arn(arn):
    """
    Validates that the ARN is for an AWS Outpost


    API Specification Document:
    https://docs.aws.amazon.com/outposts/latest/APIReference/API_Outpost.html
    """
    details = parse_aws_arn(arn)

    if not details:
        return False

    service = details.get("service") or ""
    if service.lower() != "outposts":
        return False
    resource = details.get("resource") or ""
    if not re.match("^outpost/op-[a-f0-9]{17}$", resource):
        return False

    return True



# Copyright (c) 2022 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this plugin util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import typing as t

from ansible.template import Templar


class TemplatedOptionProvider:
    def __init__(self, plugin: t.Any, templar: Templar) -> None:
        self.plugin = plugin
        self.templar = templar

    def get_option(self, option: str) -> t.Any:
        value = self.plugin.get_option(option)
        if self.templar.is_template(value):
            value = self.templar.template(variable=value)
        return value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Bruno Calogero <brunocalogero@hotmail.com>
# Copyright: (c) 2022, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_bulk_static_binding_to_epg
short_description: Bind static paths to EPGs (fv:RsPathAtt)
description:
- Bind static paths to EPGs on Cisco ACI fabrics.
options:
  tenant:
    description:
    - Name of the tenant.
    type: str
    aliases: [ tenant_name ]
  ap:
    description:
    - The name of the application profile.
    type: str
    aliases: [ app_profile, app_profile_name ]
  epg:
    description:
    - The name of the end point group.
    type: str
    aliases: [ epg_name ]
  description:
    description:
    - Description for the static path to EPG binding.
    type: str
    aliases: [ descr ]
  encap_id:
    description:
    - The encapsulation ID associating the C(epg) with the interface path.
    - This acts as the secondary C(encap_id) when using micro-segmentation.
    - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096).
    type: int
    aliases: [ vlan, vlan_id ]
  primary_encap_id:
    description:
    - Determines the primary encapsulation ID associating the C(epg)
      with the interface path when using micro-segmentation.
    - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096) and C(unknown).
    - C(unknown) is the default value and using C(unknown) disables the Micro-Segmentation.
    type: str
    aliases: [ primary_vlan, primary_vlan_id ]
  deploy_immediacy:
    description:
    - The Deployment Immediacy of Static EPG on PC, VPC or Interface.
    - The APIC defaults to C(lazy) when unset during creation.
    type: str
    choices: [ immediate, lazy ]
  interface_mode:
    description:
    - Determines how layer 2 tags will be read from and added to frames.
    - Values C(802.1p) and C(native) are identical.
    - Values C(access) and C(untagged) are identical.
    - Values C(regular), C(tagged) and C(trunk) are identical.
    - The APIC defaults to C(trunk) when unset during creation.
    type: str
    choices: [ 802.1p, access, native, regular, tagged, trunk, untagged ]
    aliases: [ interface_mode_name, mode ]
  interface_type:
    description:
    - The type of interface for the static EPG deployment.
    type: str
    choices: [ fex, port_channel, switch_port, vpc, fex_port_channel, fex_vpc ]
    default: switch_port
  interface_configs:
    description:
    - List of interface configurations, elements in the form of a dictionary.
    - Module level attributes will be overridden by the path level attributes.
    type: list
    elements: dict
    suboptions:
      description:
        description:
        - Description for the static path to EPG binding.
        type: str
        aliases: [ descr ]
      encap_id:
        description:
        - The encapsulation ID associating the C(epg) with the interface path.
        - This acts as the secondary C(encap_id) when using micro-segmentation.
        - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096).
        type: int
        aliases: [ vlan, vlan_id ]
      primary_encap_id:
        description:
        - Determines the primary encapsulation ID associating the C(epg)
          with the interface path when using micro-segmentation.
        - Accepted values are any valid encap ID for specified encap, currently ranges between C(1) and C(4096) and C(unknown).
        - C(unknown) is the default value and using C(unknown) disables the Micro-Segmentation.
        type: str
        aliases: [ primary_vlan, primary_vlan_id ]
      deploy_immediacy:
        description:
        - The Deployment Immediacy of Static EPG on PC, VPC or Interface.
        - The APIC defaults to C(lazy) when unset during creation.
        type: str
        choices: [ immediate, lazy ]
      interface_mode:
        description:
        - Determines how layer 2 tags will be read from and added to frames.
        - Values C(802.1p) and C(native) are identical.
        - Values C(access) and C(untagged) are identical.
        - Values C(regular), C(tagged) and C(trunk) are identical.
        - The APIC defaults to C(trunk) when unset during creation.
        type: str
        choices: [ 802.1p, access, native, regular, tagged, trunk, untagged ]
        aliases: [ interface_mode_name, mode ]
      interface_type:
        description:
        - The type of interface for the static EPG deployment.
        type: str
        choices: [ fex, port_channel, switch_port, vpc, fex_port_channel, fex_vpc ]
      pod_id:
        description:
        - The pod number part of the tDn.
        - C(pod_id) is usually an integer below C(10).
        type: int
        required: true
        aliases: [ pod, pod_number ]
      leafs:
        description:
        - The switch ID(s) that the C(interface) belongs to.
        - When C(interface_type) is C(switch_port), C(port_channel), or C(fex), then C(leafs) is a string of the leaf ID.
        - When C(interface_type) is C(vpc), then C(leafs) is a list with both leaf IDs.
        - The C(leafs) value is usually something like '101' or '101-102' depending on C(connection_type).
        type: list
        elements: str
        required: true
        aliases: [ leaves, nodes, paths, switches ]
      interface:
        description:
        - The C(interface) string value part of the tDn.
        - Usually a policy group like C(test-IntPolGrp) or an interface of the following format C(1/7) depending on C(interface_type).
        type: str
        required: true
      extpaths:
        description:
        - The C(extpaths) integer value part of the tDn.
        - C(extpaths) is only used if C(interface_type) is C(fex), C(fex_vpc) or C(fex_port_channel).
        - When C(interface_type) is C(fex_vpc), then C(extpaths) is a list with both fex IDs.
        - Usually something like C(1011).
        type: list
        elements: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant), C(ap), C(epg) used must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant), M(cisco.aci.aci_ap), M(cisco.aci.aci_epg) modules can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- module: cisco.aci.aci_ap
- module: cisco.aci.aci_epg
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(fv:RsPathAtt).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Bruno Calogero (@brunocalogero)
- Marcel Zehnder (@maercu)
- Sabari Jaganathan (@sajagana)
"""

EXAMPLES = r"""
- name: Create list of interfaces using module level attributes
  cisco.aci.aci_bulk_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    encap_id: 221
    interface_mode: trunk
    deploy_immediacy: lazy
    description: "Module level attributes used to create interfaces"
    interface_configs:
      - interface: 1/7
        leafs: 101
        pod: 1
      - interface: 1/7
        leafs: 107
        pod: 7
      - interface: 1/8
        leafs: 108
        pod: 8
        encap_id: 108
    state: present
  delegate_to: localhost

- name: Create/Update list of interfaces using path level attributes
  cisco.aci.aci_bulk_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    interface_configs:
      - interface: 1/7
        leafs: 101
        pod: 1
        encap_id: 221
        interface_mode: trunk
        deploy_immediacy: lazy
        description: "Path level attributes used to create/update interfaces"
      - interface: 1/7
        leafs: 107
        pod: 7
        encap_id: 221
        interface_mode: trunk
        deploy_immediacy: lazy
        description: "Path level attributes used to create/update interfaces"
      - interface: 1/8
        leafs: 108
        pod: 8
        encap_id: 108
        interface_mode: trunk
        deploy_immediacy: lazy
        description: "Path level attributes used to create/update interfaces"
    state: present
  delegate_to: localhost

- name: Query all interfaces of an EPG
  cisco.aci.aci_bulk_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    state: query
  delegate_to: localhost

- name: Query all interfaces
  cisco.aci.aci_bulk_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost

- name: Remove list of interfaces
  cisco.aci.aci_bulk_static_binding_to_epg:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: accessport-code-cert
    ap: accessport_code_app
    epg: accessport_epg1
    encap_id: 221
    interface_mode: trunk
    deploy_immediacy: lazy
    interface_configs:
      - interface: 1/7
        leafs: 101
        pod: 1
      - interface: 1/7
        leafs: 107
        pod: 7
      - interface: 1/8
        leafs: 108
        pod: 8
        encap_id: 108
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec

INTERFACE_MODE_MAPPING = {
    "802.1p": "native",
    "access": "untagged",
    "native": "native",
    "regular": "regular",
    "tagged": "regular",
    "trunk": "regular",
    "untagged": "untagged",
}

INTERFACE_TYPE_MAPPING = {
    "fex": "topology/pod-{pod_id}/paths-{leafs}/extpaths-{extpaths}/pathep-[eth{interface}]",
    "fex_port_channel": "topology/pod-{pod_id}/paths-{leafs}/extpaths-{extpaths}/pathep-[{interface}]",
    "fex_vpc": "topology/pod-{pod_id}/protpaths-{leafs}/extprotpaths-{extpaths}/pathep-[{interface}]",
    "port_channel": "topology/pod-{pod_id}/paths-{leafs}/pathep-[{interface}]",
    "switch_port": "topology/pod-{pod_id}/paths-{leafs}/pathep-[eth{interface}]",
    "vpc": "topology/pod-{pod_id}/protpaths-{leafs}/pathep-[{interface}]",
}

INTERFACE_STATUS_MAPPING = {"absent": "deleted"}


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),
        ap=dict(type="str", aliases=["app_profile", "app_profile_name"]),
        epg=dict(type="str", aliases=["epg_name"]),
        description=dict(type="str", aliases=["descr"]),
        encap_id=dict(type="int", aliases=["vlan", "vlan_id"]),
        primary_encap_id=dict(type="str", aliases=["primary_vlan", "primary_vlan_id"]),
        deploy_immediacy=dict(type="str", choices=["immediate", "lazy"]),
        interface_mode=dict(
            type="str", choices=["802.1p", "access", "native", "regular", "tagged", "trunk", "untagged"], aliases=["interface_mode_name", "mode"]
        ),
        interface_type=dict(type="str", default="switch_port", choices=["fex", "port_channel", "switch_port", "vpc", "fex_port_channel", "fex_vpc"]),
        interface_configs=dict(
            type="list",
            elements="dict",
            options=dict(
                description=dict(type="str", aliases=["descr"]),
                encap_id=dict(type="int", aliases=["vlan", "vlan_id"]),
                primary_encap_id=dict(type="str", aliases=["primary_vlan", "primary_vlan_id"]),
                deploy_immediacy=dict(type="str", choices=["immediate", "lazy"]),
                interface_mode=dict(
                    type="str", choices=["802.1p", "access", "native", "regular", "tagged", "trunk", "untagged"], aliases=["interface_mode_name", "mode"]
                ),
                interface_type=dict(type="str", choices=["fex", "port_channel", "switch_port", "vpc", "fex_port_channel", "fex_vpc"]),
                pod_id=dict(type="int", required=True, aliases=["pod", "pod_number"]),
                leafs=dict(type="list", elements="str", required=True, aliases=["leaves", "nodes", "paths", "switches"]),
                interface=dict(type="str", required=True),
                extpaths=dict(type="list", elements="str"),
            ),
        ),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["ap", "epg", "tenant", "interface_configs"]],
            ["state", "present", ["ap", "epg", "tenant", "interface_configs"]],
        ],
    )

    tenant = module.params.get("tenant")
    ap = module.params.get("ap")
    epg = module.params.get("epg")
    module_description = module.params.get("description")
    module_encap_id = module.params.get("encap_id")
    module_primary_encap_id = module.params.get("primary_encap_id")
    module_deploy_immediacy = module.params.get("deploy_immediacy")
    module_interface_mode = module.params.get("interface_mode")
    module_interface_type = module.params.get("interface_type")
    interface_configs = module.params.get("interface_configs")
    state = module.params.get("state")

    aci = ACIModule(module)
    children = []

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter=dict(name=tenant),
        ),
        subclass_1=dict(
            aci_class="fvAp",
            aci_rn="ap-{0}".format(ap),
            module_object=ap,
            target_filter=dict(name=ap),
        ),
        subclass_2=dict(
            aci_class="fvAEPg",
            aci_rn="epg-{0}".format(epg),
            module_object=epg,
            target_filter=dict(name=epg),
        ),
        child_classes=["fvRsPathAtt"],
    )

    aci.get_existing()

    if state == "present" or state == "absent":
        for interface_config in interface_configs:
            pod_id = interface_config.get("pod_id")
            interface = interface_config.get("interface")
            extpaths = interface_config.get("extpaths")

            description = interface_config.get("description") or module_description
            deploy_immediacy = interface_config.get("deploy_immediacy") or module_deploy_immediacy
            interface_type = interface_config.get("interface_type") or module_interface_type
            encap_id = interface_config.get("encap_id") or module_encap_id
            primary_encap_id = interface_config.get("primary_encap_id") or module_primary_encap_id
            interface_mode = interface_config.get("interface_mode") or module_interface_mode

            if interface_type in ["fex", "fex_vpc", "fex_port_channel"] and extpaths is None:
                aci.fail_json(msg="extpaths is required when interface_type is: {0}".format(interface_type))

            # Process leafs, and support dash-delimited leafs
            leafs = []
            for leaf in interface_config.get("leafs"):
                # Users are likely to use integers for leaf IDs, which would raise an exception when using the join method
                leafs.extend(str(leaf).split("-"))
            if len(leafs) == 1:
                if interface_type in ["vpc", "fex_vpc"]:
                    aci.fail_json(msg='A interface_type of "vpc" requires 2 leafs')
                leafs = leafs[0]
            elif len(leafs) == 2:
                if interface_type not in ["vpc", "fex_vpc"]:
                    aci.fail_json(msg='The interface_types "switch_port", "port_channel", and "fex" do not support using multiple leafs for a single binding')
                leafs = "-".join(leafs)
            else:
                aci.fail_json(msg='The "leafs" parameter must not have more than 2 entries')

            if extpaths is not None:
                # Process extpaths, and support dash-delimited extpaths
                extpaths = []
                for extpath in interface_config.get("extpaths"):
                    # Users are likely to use integers for extpaths IDs, which would raise an exception when using the join method
                    extpaths.extend(str(extpath).split("-"))
                if len(extpaths) == 1:
                    if interface_type == "fex_vpc":
                        aci.fail_json(msg='A interface_type of "fex_vpc" requires 2 extpaths')
                    extpaths = extpaths[0]
                elif len(extpaths) == 2:
                    if interface_type != "fex_vpc":
                        aci.fail_json(msg='The interface_types "fex" and "fex_port_channel" do not support using multiple extpaths for a single binding')
                    extpaths = "-".join(extpaths)
                else:
                    aci.fail_json(msg='The "extpaths" parameter must not have more than 2 entries')

            if encap_id is not None:
                if encap_id not in range(1, 4097):
                    aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096")
                encap_id = "vlan-{0}".format(encap_id)

            if primary_encap_id is not None:
                try:
                    primary_encap_id = int(primary_encap_id)
                    if isinstance(primary_encap_id, int) and primary_encap_id in range(1, 4097):
                        primary_encap_id = "vlan-{0}".format(primary_encap_id)
                    else:
                        aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096 or unknown.")
                except Exception as e:
                    if isinstance(primary_encap_id, str) and primary_encap_id != "unknown":
                        aci.fail_json(msg="Valid VLAN assignments are from 1 to 4096 or unknown. %s" % e)

            static_path = INTERFACE_TYPE_MAPPING[interface_type].format(pod_id=pod_id, leafs=leafs, extpaths=extpaths, interface=interface)

            interface_mode = INTERFACE_MODE_MAPPING.get(interface_mode)

            interface_status = INTERFACE_STATUS_MAPPING.get(state)

            children.append(
                dict(
                    fvRsPathAtt=dict(
                        attributes=dict(
                            descr=description,
                            encap=encap_id,
                            primaryEncap=primary_encap_id,
                            instrImedcy=deploy_immediacy,
                            mode=interface_mode,
                            tDn=static_path,
                            status=interface_status,
                        )
                    )
                )
            )

        aci.payload(
            aci_class="fvAEPg",
            class_config=dict(),
            child_configs=children,
        )

        aci.get_diff(aci_class="fvAEPg")

        aci.post_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The parse_xml plugin code
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import os
import traceback

from collections.abc import Mapping
from xml.etree.ElementTree import fromstring

from ansible.errors import AnsibleFilterError
from ansible.module_utils.common.text.converters import to_native
from ansible.utils.display import Display

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    JinjaTemplate,
)


string_types = (str,)

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

display = Display()


def _raise_error(msg):
    raise AnsibleFilterError(msg)


def _extract_param(template, root, attrs, value):
    key = None
    when = attrs.get("when")
    conditional = "{%% if %s %%}True{%% else %%}False{%% endif %%}" % when
    param_to_xpath_map = attrs["items"]

    if isinstance(value, Mapping):
        key = value.get("key", None)
        if key:
            value = value["values"]

    entries = dict() if key else list()

    for element in root.findall(attrs["top"]):
        entry = dict()
        item_dict = dict()
        for param, param_xpath in param_to_xpath_map.items():
            fields = None
            try:
                fields = element.findall(param_xpath)
            except Exception:
                display.warning(
                    "Failed to evaluate value of '%s' with XPath '%s'.\nUnexpected error: %s."
                    % (param, param_xpath, traceback.format_exc())
                )

            tags = param_xpath.split("/")

            # check if xpath ends with attribute.
            # If yes set attribute key/value dict to param value in case attribute matches
            # else if it is a normal xpath assign matched element text value.
            if len(tags) and tags[-1].endswith("]"):
                if fields:
                    if len(fields) > 1:
                        item_dict[param] = [field.attrib for field in fields]
                    else:
                        item_dict[param] = fields[0].attrib
                else:
                    item_dict[param] = {}
            else:
                if fields:
                    if len(fields) > 1:
                        item_dict[param] = [field.text for field in fields]
                    else:
                        item_dict[param] = fields[0].text
                else:
                    item_dict[param] = None

        if isinstance(value, Mapping):
            for item_key, item_value in value.items():
                entry[item_key] = template(item_value, {"item": item_dict})
        else:
            entry = template(value, {"item": item_dict})

        if key:
            expanded_key = template(key, {"item": item_dict})
            if when:
                if template(
                    conditional,
                    {"item": {"key": expanded_key, "value": entry}},
                ):
                    entries[expanded_key] = entry
            else:
                entries[expanded_key] = entry
        else:
            if when:
                if template(conditional, {"item": entry}):
                    entries.append(entry)
            else:
                entries.append(entry)

    return entries


def parse_xml(output, tmpl):
    if not os.path.exists(tmpl):
        _raise_error("unable to locate parse_xml template: %s" % tmpl)

    if not isinstance(output, string_types):
        _raise_error("parse_xml works on string input, but given input of : %s" % type(output))

    root = fromstring(output)
    try:
        template = JinjaTemplate()
    except ImportError as exc:
        raise AnsibleFilterError(to_native(exc))

    with open(tmpl) as tmpl_fh:
        tmpl_content = tmpl_fh.read()

    spec = yaml.safe_load(tmpl_content)
    obj = {}

    for name, attrs in spec["keys"].items():
        value = attrs["value"]

        try:
            variables = spec.get("vars", {})
            value = template(value, variables)
        except Exception:
            pass

        if "items" in attrs:
            obj[name] = _extract_param(template, root, attrs, value)
        else:
            obj[name] = value

    return obj



# Copyright (c) 2020 FERREIRA Christophe <christophe.ferreira@cnaf.fr>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Standard files documentation fragment
    DOCUMENTATION = r"""
options:
  host:
    description:
      - FQDN of Nomad server.
    required: true
    type: str
  port:
    description:
      - Port of Nomad server.
    type: int
    default: 4646
    version_added: 8.0.0
  use_ssl:
    description:
      - Use TLS/SSL connection.
    type: bool
    default: true
  timeout:
    description:
      - Timeout (in seconds) for the request to Nomad.
    type: int
    default: 5
  validate_certs:
    description:
      - Enable TLS/SSL certificate validation.
    type: bool
    default: true
  client_cert:
    description:
      - Path of certificate for TLS/SSL.
    type: path
  client_key:
    description:
      - Path of certificate's private key for TLS/SSL.
    type: path
  namespace:
    description:
      - Namespace for Nomad.
    type: str
  token:
    description:
      - ACL token for authentication.
    type: str
"""



# Copyright (c) 2021, Andrew Pantuso (@ajpantuso) <ajpantuso@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import abc
import binascii
import datetime as _datetime
import os
import typing as t
from base64 import b64encode
from datetime import datetime
from hashlib import sha256

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.crypto.plugins.module_utils._openssh.utils import (
    OpensshParser,
    _OpensshWriter,
)
from ansible_collections.community.crypto.plugins.module_utils._time import UTC as _UTC
from ansible_collections.community.crypto.plugins.module_utils._time import (
    add_or_remove_timezone as _add_or_remove_timezone,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    convert_relative_to_datetime,
)

if t.TYPE_CHECKING:
    from ansible_collections.community.crypto.plugins.module_utils._openssh.cryptography import (  # pragma: no cover
        PubKeyType,
    )

    DateFormat = t.Literal["human_readable", "openssh", "timestamp"]  # pragma: no cover
    DateFormatStr = t.Literal["human_readable", "openssh"]  # pragma: no cover
    DateFormatInt = t.Literal["timestamp"]  # pragma: no cover


# Protocol References
# -------------------
# https://datatracker.ietf.org/doc/html/rfc4251
# https://datatracker.ietf.org/doc/html/rfc4253
# https://datatracker.ietf.org/doc/html/rfc5656
# https://datatracker.ietf.org/doc/html/rfc8032
# https://cvsweb.openbsd.org/src/usr.bin/ssh/PROTOCOL.certkeys?annotate=HEAD
#
# Inspired by:
# ------------
# https://github.com/pyca/cryptography/blob/main/src/cryptography/hazmat/primitives/serialization/ssh.py
# https://github.com/paramiko/paramiko/blob/master/paramiko/message.py


# See https://cvsweb.openbsd.org/src/usr.bin/ssh/PROTOCOL.certkeys?annotate=HEAD
_USER_TYPE = 1
_HOST_TYPE = 2

_SSH_TYPE_STRINGS: dict[PubKeyType, bytes] = {
    "rsa": b"ssh-rsa",
    "dsa": b"ssh-dss",
    "ecdsa-nistp256": b"ecdsa-sha2-nistp256",
    "ecdsa-nistp384": b"ecdsa-sha2-nistp384",
    "ecdsa-nistp521": b"ecdsa-sha2-nistp521",
    "ed25519": b"ssh-ed25519",
}
_CERT_SUFFIX_V01 = b"-cert-v01@openssh.com"

# See https://datatracker.ietf.org/doc/html/rfc5656#section-6.1
_ECDSA_CURVE_IDENTIFIERS = {
    "ecdsa-nistp256": b"nistp256",
    "ecdsa-nistp384": b"nistp384",
    "ecdsa-nistp521": b"nistp521",
}
_ECDSA_CURVE_IDENTIFIERS_LOOKUP: dict[bytes, PubKeyType] = {
    b"nistp256": "ecdsa-nistp256",
    b"nistp384": "ecdsa-nistp384",
    b"nistp521": "ecdsa-nistp521",
}

_ALWAYS = _add_or_remove_timezone(datetime(1970, 1, 1), with_timezone=True)
_FOREVER = datetime(9999, 12, 31, 23, 59, 59, 999999, _UTC)

_CRITICAL_OPTIONS = (
    "force-command",
    "source-address",
    "verify-required",
)

_DIRECTIVES = (
    "clear",
    "no-x11-forwarding",
    "no-agent-forwarding",
    "no-port-forwarding",
    "no-pty",
    "no-user-rc",
)

_EXTENSIONS = (
    "permit-x11-forwarding",
    "permit-agent-forwarding",
    "permit-port-forwarding",
    "permit-pty",
    "permit-user-rc",
)


class OpensshCertificateTimeParameters:
    def __init__(
        self, *, valid_from: str | bytes | int, valid_to: str | bytes | int
    ) -> None:
        self._valid_from = self.to_datetime(valid_from)
        self._valid_to = self.to_datetime(valid_to)

        if self._valid_from > self._valid_to:
            raise ValueError(
                f"Valid from: {valid_from!r} must not be greater than Valid to: {valid_to!r}"
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self._valid_from == other._valid_from and self._valid_to == other._valid_to
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    @property
    def validity_string(self) -> str:
        if not (self._valid_from == _ALWAYS and self._valid_to == _FOREVER):
            return f"{self.valid_from(date_format='openssh')}:{self.valid_to(date_format='openssh')}"
        return ""

    @t.overload
    def valid_from(self, date_format: DateFormatStr) -> str: ...

    @t.overload
    def valid_from(self, date_format: DateFormatInt) -> int: ...

    @t.overload
    def valid_from(self, date_format: DateFormat) -> str | int: ...

    def valid_from(self, date_format: DateFormat) -> str | int:
        return self.format_datetime(self._valid_from, date_format=date_format)

    @t.overload
    def valid_to(self, date_format: DateFormatStr) -> str: ...

    @t.overload
    def valid_to(self, date_format: DateFormatInt) -> int: ...

    @t.overload
    def valid_to(self, date_format: DateFormat) -> str | int: ...

    def valid_to(self, date_format: DateFormat) -> str | int:
        return self.format_datetime(self._valid_to, date_format=date_format)

    def within_range(self, valid_at: str | bytes | int | None) -> bool:
        if valid_at is not None:
            valid_at_datetime = self.to_datetime(valid_at)
            return self._valid_from <= valid_at_datetime <= self._valid_to
        return True

    @t.overload
    @staticmethod
    def format_datetime(dt: datetime, *, date_format: DateFormatStr) -> str: ...

    @t.overload
    @staticmethod
    def format_datetime(dt: datetime, *, date_format: DateFormatInt) -> int: ...

    @t.overload
    @staticmethod
    def format_datetime(dt: datetime, *, date_format: DateFormat) -> str | int: ...

    @staticmethod
    def format_datetime(dt: datetime, *, date_format: DateFormat) -> str | int:
        if date_format in ("human_readable", "openssh"):
            if dt == _ALWAYS:
                return "always"
            if dt == _FOREVER:
                return "forever"
            return (
                dt.isoformat().replace("+00:00", "")
                if date_format == "human_readable"
                else dt.strftime("%Y%m%d%H%M%S")
            )
        if date_format == "timestamp":
            td = dt - _ALWAYS
            return int(
                (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
            )
        raise ValueError(f"{date_format} is not a valid format")

    @staticmethod
    def to_datetime(time_string_or_timestamp: str | bytes | int) -> datetime:
        if isinstance(time_string_or_timestamp, (str, bytes)):
            return OpensshCertificateTimeParameters._time_string_to_datetime(
                to_text(time_string_or_timestamp.strip())
            )
        if isinstance(time_string_or_timestamp, int):
            return OpensshCertificateTimeParameters._timestamp_to_datetime(
                time_string_or_timestamp
            )
        raise ValueError(
            f"Value must be of type (str, unicode, int) not {type(time_string_or_timestamp)}"
        )

    @staticmethod
    def _timestamp_to_datetime(timestamp: int) -> datetime:
        if timestamp == 0x0:
            return _ALWAYS
        if timestamp == 0xFFFFFFFFFFFFFFFF:
            return _FOREVER
        try:
            return datetime.fromtimestamp(timestamp, tz=_datetime.timezone.utc)
        except OverflowError as e:
            raise ValueError from e

    @staticmethod
    def _time_string_to_datetime(time_string: str) -> datetime:
        if time_string == "always":
            return _ALWAYS
        if time_string == "forever":
            return _FOREVER
        if is_relative_time_string(time_string):
            result = convert_relative_to_datetime(time_string, with_timezone=True)
            if result is None:
                raise ValueError
            return result
        result = None
        for time_format in ("%Y-%m-%d", "%Y-%m-%d %H:%M:%S", "%Y-%m-%dT%H:%M:%S"):
            try:
                result = _add_or_remove_timezone(
                    datetime.strptime(time_string, time_format),
                    with_timezone=True,
                )
            except ValueError:
                pass
        if result is None:
            raise ValueError
        return result


_OpensshCertificateOption = t.TypeVar(
    "_OpensshCertificateOption", bound="OpensshCertificateOption"
)


class OpensshCertificateOption:
    def __init__(
        self,
        *,
        option_type: t.Literal["critical", "extension"],
        name: str | bytes,
        data: str | bytes,
    ):
        if option_type not in ("critical", "extension"):
            raise ValueError("type must be either 'critical' or 'extension'")

        if not isinstance(name, (str, bytes)):
            raise TypeError(f"name must be a string not {type(name)}")

        if not isinstance(data, (str, bytes)):
            raise TypeError(f"data must be a string not {type(data)}")

        self._option_type = option_type
        self._name = name.lower()
        self._data = data

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented

        return all(
            [
                self._option_type == other._option_type,
                self._name == other._name,
                self._data == other._data,
            ]
        )

    def __hash__(self) -> int:
        return hash((self._option_type, self._name, self._data))

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __str__(self) -> str:
        if self._data:
            return f"{self._name!r}={self._data!r}"
        return f"{self._name!r}"

    @property
    def data(self) -> str | bytes:
        return self._data

    @property
    def name(self) -> str | bytes:
        return self._name

    @property
    def type(self) -> t.Literal["critical", "extension"]:
        return self._option_type

    @classmethod
    def from_string(
        cls: t.Type[_OpensshCertificateOption], option_string: str  # noqa: UP006
    ) -> _OpensshCertificateOption:
        if not isinstance(option_string, str):
            raise ValueError(
                f"option_string must be a string not {type(option_string)}"
            )
        option_type = None

        if ":" in option_string:
            option_type, value = option_string.strip().split(":", 1)
            if "=" in value:
                name, data = value.split("=", 1)
            else:
                name, data = value, ""
        elif "=" in option_string:
            name, data = option_string.strip().split("=", 1)
        else:
            name, data = option_string.strip(), ""

        return cls(
            # We have str, but we're expecting a specific literal:
            option_type=option_type or get_option_type(name.lower()),  # type: ignore
            name=name,
            data=data,
        )


if t.TYPE_CHECKING:

    class _OpensshCertificateInfoKwarg(t.TypedDict):
        nonce: t.NotRequired[bytes | None]
        serial: t.NotRequired[int | None]
        cert_type: t.NotRequired[int | None]
        key_id: t.NotRequired[bytes | None]
        principals: t.NotRequired[list[bytes] | None]
        valid_after: t.NotRequired[int | None]
        valid_before: t.NotRequired[int | None]
        critical_options: t.NotRequired[list[tuple[bytes, bytes]] | None]
        extensions: t.NotRequired[list[tuple[bytes, bytes]] | None]
        reserved: t.NotRequired[bytes | None]
        signing_key: t.NotRequired[bytes | None]


class OpensshCertificateInfo(metaclass=abc.ABCMeta):
    """Encapsulates all certificate information which is signed by a CA key"""

    def __init__(
        self,
        *,
        nonce: bytes | None = None,
        serial: int | None = None,
        cert_type: int | None = None,
        key_id: bytes | None = None,
        principals: list[bytes] | None = None,
        valid_after: int | None = None,
        valid_before: int | None = None,
        critical_options: list[tuple[bytes, bytes]] | None = None,
        extensions: list[tuple[bytes, bytes]] | None = None,
        reserved: bytes | None = None,
        signing_key: bytes | None = None,
    ):
        self.nonce = nonce
        self.serial = serial
        self._cert_type: int | None = cert_type
        self.key_id = key_id
        self.principals = principals
        self.valid_after = valid_after
        self.valid_before = valid_before
        self.critical_options = critical_options
        self.extensions = extensions
        self.reserved = reserved
        self.signing_key = signing_key

        self.type_string: bytes | None = None

    @property
    def cert_type(self) -> t.Literal["user", "host", ""]:
        if self._cert_type == _USER_TYPE:
            return "user"
        if self._cert_type == _HOST_TYPE:
            return "host"
        return ""

    @cert_type.setter
    def cert_type(self, cert_type: t.Literal["user", "host"] | int) -> None:
        if cert_type in ("user", _USER_TYPE):
            self._cert_type = _USER_TYPE
        elif cert_type in ("host", _HOST_TYPE):
            self._cert_type = _HOST_TYPE
        else:
            raise ValueError(f"{cert_type} is not a valid certificate type")

    def signing_key_fingerprint(self) -> bytes:
        if self.signing_key is None:
            raise ValueError("signing_key not present")
        return fingerprint(self.signing_key)

    @abc.abstractmethod
    def public_key_fingerprint(self) -> bytes:
        pass

    @abc.abstractmethod
    def parse_public_numbers(self, parser: OpensshParser) -> None:
        pass


class OpensshRSACertificateInfo(OpensshCertificateInfo):
    def __init__(
        self,
        *,
        e: int | None = None,
        n: int | None = None,
        **kwargs: t.Unpack[_OpensshCertificateInfoKwarg],
    ) -> None:
        super().__init__(**kwargs)
        self.type_string = _SSH_TYPE_STRINGS["rsa"] + _CERT_SUFFIX_V01
        self.e = e
        self.n = n

    # See https://datatracker.ietf.org/doc/html/rfc4253#section-6.6
    def public_key_fingerprint(self) -> bytes:
        if self.e is None or self.n is None:
            return b""

        writer = _OpensshWriter()
        writer.string(_SSH_TYPE_STRINGS["rsa"])
        writer.mpint(self.e)
        writer.mpint(self.n)

        return fingerprint(writer.bytes())

    def parse_public_numbers(self, parser: OpensshParser) -> None:
        self.e = parser.mpint()
        self.n = parser.mpint()


class OpensshDSACertificateInfo(OpensshCertificateInfo):
    def __init__(
        self,
        *,
        p: int | None = None,
        q: int | None = None,
        g: int | None = None,
        y: int | None = None,
        **kwargs: t.Unpack[_OpensshCertificateInfoKwarg],
    ) -> None:
        super().__init__(**kwargs)
        self.type_string = _SSH_TYPE_STRINGS["dsa"] + _CERT_SUFFIX_V01
        self.p = p
        self.q = q
        self.g = g
        self.y = y

    # See https://datatracker.ietf.org/doc/html/rfc4253#section-6.6
    def public_key_fingerprint(self) -> bytes:
        if self.p is None or self.q is None or self.g is None or self.y is None:
            return b""

        writer = _OpensshWriter()
        writer.string(_SSH_TYPE_STRINGS["dsa"])
        writer.mpint(self.p)
        writer.mpint(self.q)
        writer.mpint(self.g)
        writer.mpint(self.y)

        return fingerprint(writer.bytes())

    def parse_public_numbers(self, parser: OpensshParser) -> None:
        self.p = parser.mpint()
        self.q = parser.mpint()
        self.g = parser.mpint()
        self.y = parser.mpint()


class OpensshECDSACertificateInfo(OpensshCertificateInfo):
    def __init__(
        self,
        *,
        curve: bytes | None = None,
        public_key: bytes | None = None,
        **kwargs: t.Unpack[_OpensshCertificateInfoKwarg],
    ):
        super().__init__(**kwargs)
        self._curve: bytes | None = None
        if curve is not None:
            self.curve = curve

        self.public_key = public_key

    @property
    def curve(self) -> bytes | None:
        return self._curve

    @curve.setter
    def curve(self, curve: bytes) -> None:
        if curve in _ECDSA_CURVE_IDENTIFIERS.values():
            self._curve = curve
            self.type_string = (
                _SSH_TYPE_STRINGS[_ECDSA_CURVE_IDENTIFIERS_LOOKUP[curve]]
                + _CERT_SUFFIX_V01
            )
        else:
            raise ValueError(
                "Curve must be one of {(b','.join(_ECDSA_CURVE_IDENTIFIERS.values())).decode('UTF-8')}"
            )

    # See https://datatracker.ietf.org/doc/html/rfc4253#section-6.6
    def public_key_fingerprint(self) -> bytes:
        if self.curve is None or self.public_key is None:
            return b""

        writer = _OpensshWriter()
        writer.string(_SSH_TYPE_STRINGS[_ECDSA_CURVE_IDENTIFIERS_LOOKUP[self.curve]])
        writer.string(self.curve)
        writer.string(self.public_key)

        return fingerprint(writer.bytes())

    def parse_public_numbers(self, parser: OpensshParser) -> None:
        self.curve = parser.string()
        self.public_key = parser.string()


class OpensshED25519CertificateInfo(OpensshCertificateInfo):
    def __init__(
        self,
        *,
        pk: bytes | None = None,
        **kwargs: t.Unpack[_OpensshCertificateInfoKwarg],
    ) -> None:
        super().__init__(**kwargs)
        self.type_string = _SSH_TYPE_STRINGS["ed25519"] + _CERT_SUFFIX_V01
        self.pk = pk

    def public_key_fingerprint(self) -> bytes:
        if self.pk is None:
            return b""

        writer = _OpensshWriter()
        writer.string(_SSH_TYPE_STRINGS["ed25519"])
        writer.string(self.pk)

        return fingerprint(writer.bytes())

    def parse_public_numbers(self, parser: OpensshParser) -> None:
        self.pk = parser.string()


_OpensshCertificate = t.TypeVar("_OpensshCertificate", bound="OpensshCertificate")


# See https://cvsweb.openbsd.org/src/usr.bin/ssh/PROTOCOL.certkeys?annotate=HEAD
class OpensshCertificate:
    """Encapsulates a formatted OpenSSH certificate including signature and signing key"""

    def __init__(self, *, cert_info: OpensshCertificateInfo, signature: bytes):
        self._cert_info = cert_info
        self.signature = signature

    @classmethod
    def load(
        cls: t.Type[_OpensshCertificate], path: str | os.PathLike  # noqa: UP006
    ) -> _OpensshCertificate:
        if not os.path.exists(path):
            raise ValueError(f"{path} is not a valid path.")

        try:
            with open(path, "rb") as cert_file:
                data = cert_file.read()
        except (IOError, OSError) as e:
            raise ValueError(f"{path} cannot be opened for reading: {e}") from e

        try:
            format_identifier, b64_cert = data.split(b" ")[:2]
            cert = binascii.a2b_base64(b64_cert)
        except (binascii.Error, ValueError) as e:
            raise ValueError("Certificate not in OpenSSH format") from e

        for key_type, string in _SSH_TYPE_STRINGS.items():
            if format_identifier == string + _CERT_SUFFIX_V01:
                pub_key_type = key_type
                break
        else:
            raise ValueError(
                f"Invalid certificate format identifier: {format_identifier!r}"
            )

        parser = OpensshParser(data=cert)

        if format_identifier != parser.string():
            raise ValueError("Certificate formats do not match")

        try:
            cert_info = cls._parse_cert_info(pub_key_type, parser)
            signature = parser.string()
        except (TypeError, ValueError) as e:
            raise ValueError(f"Invalid certificate data: {e}") from e

        if parser.remaining_bytes():
            raise ValueError(
                f"{parser.remaining_bytes()} bytes of additional data was not parsed while loading {path}"
            )

        return cls(
            cert_info=cert_info,
            signature=signature,
        )

    @property
    def type_string(self) -> str:
        return to_text(self._cert_info.type_string)

    @property
    def nonce(self) -> bytes:
        if self._cert_info.nonce is None:
            raise ValueError
        return self._cert_info.nonce

    @property
    def public_key(self) -> str:
        return to_text(self._cert_info.public_key_fingerprint())

    @property
    def serial(self) -> int:
        if self._cert_info.serial is None:
            raise ValueError
        return self._cert_info.serial

    @property
    def type(self) -> t.Literal["user", "host"]:
        result = self._cert_info.cert_type
        if result == "":
            raise ValueError
        return result

    @property
    def key_id(self) -> str:
        return to_text(self._cert_info.key_id)

    @property
    def principals(self) -> list[str]:
        if self._cert_info.principals is None:
            raise ValueError
        return [to_text(p) for p in self._cert_info.principals]

    @property
    def valid_after(self) -> int:
        if self._cert_info.valid_after is None:
            raise ValueError
        return self._cert_info.valid_after

    @property
    def valid_before(self) -> int:
        if self._cert_info.valid_before is None:
            raise ValueError
        return self._cert_info.valid_before

    @property
    def critical_options(self) -> list[OpensshCertificateOption]:
        if self._cert_info.critical_options is None:
            raise ValueError
        return [
            OpensshCertificateOption(
                option_type="critical", name=to_text(n), data=to_text(d)
            )
            for n, d in self._cert_info.critical_options
        ]

    @property
    def extensions(self) -> list[OpensshCertificateOption]:
        if self._cert_info.extensions is None:
            raise ValueError
        return [
            OpensshCertificateOption(
                option_type="extension", name=to_text(n), data=to_text(d)
            )
            for n, d in self._cert_info.extensions
        ]

    @property
    def reserved(self) -> bytes:
        if self._cert_info.reserved is None:
            raise ValueError
        return self._cert_info.reserved

    @property
    def signing_key(self) -> str:
        return to_text(self._cert_info.signing_key_fingerprint())

    @property
    def signature_type(self) -> str:
        signature_data = OpensshParser.signature_data(signature_string=self.signature)
        return to_text(signature_data["signature_type"])

    @staticmethod
    def _parse_cert_info(
        pub_key_type: PubKeyType, parser: OpensshParser
    ) -> OpensshCertificateInfo:
        cert_info = get_cert_info_object(pub_key_type)
        cert_info.nonce = parser.string()
        cert_info.parse_public_numbers(parser)
        cert_info.serial = parser.uint64()
        # mypy doesn't understand that the setter accepts other types than the getter:
        cert_info.cert_type = parser.uint32()  # type: ignore
        cert_info.key_id = parser.string()
        cert_info.principals = parser.string_list()
        cert_info.valid_after = parser.uint64()
        cert_info.valid_before = parser.uint64()
        cert_info.critical_options = parser.option_list()
        cert_info.extensions = parser.option_list()
        cert_info.reserved = parser.string()
        cert_info.signing_key = parser.string()

        return cert_info

    def to_dict(self) -> dict[str, t.Any]:
        time_parameters = OpensshCertificateTimeParameters(
            valid_from=self.valid_after, valid_to=self.valid_before
        )
        return {
            "type_string": self.type_string,
            "nonce": self.nonce,
            "serial": self.serial,
            "cert_type": self.type,
            "identifier": self.key_id,
            "principals": self.principals,
            "valid_after": time_parameters.valid_from(date_format="human_readable"),
            "valid_before": time_parameters.valid_to(date_format="human_readable"),
            "critical_options": [
                str(critical_option) for critical_option in self.critical_options
            ],
            "extensions": [str(extension) for extension in self.extensions],
            "reserved": self.reserved,
            "public_key": self.public_key,
            "signing_key": self.signing_key,
        }


def apply_directives(directives: t.Iterable[str]) -> list[OpensshCertificateOption]:
    if any(d not in _DIRECTIVES for d in directives):
        raise ValueError(f"directives must be one of {', '.join(_DIRECTIVES)}")

    directive_to_option = {
        "no-x11-forwarding": OpensshCertificateOption(
            option_type="extension", name="permit-x11-forwarding", data=""
        ),
        "no-agent-forwarding": OpensshCertificateOption(
            option_type="extension", name="permit-agent-forwarding", data=""
        ),
        "no-port-forwarding": OpensshCertificateOption(
            option_type="extension", name="permit-port-forwarding", data=""
        ),
        "no-pty": OpensshCertificateOption(
            option_type="extension", name="permit-pty", data=""
        ),
        "no-user-rc": OpensshCertificateOption(
            option_type="extension", name="permit-user-rc", data=""
        ),
    }

    if "clear" in directives:
        return []
    return list(
        set(default_options()) - set(directive_to_option[d] for d in directives)
    )


def default_options() -> list[OpensshCertificateOption]:
    return [
        OpensshCertificateOption(option_type="extension", name=name, data="")
        for name in _EXTENSIONS
    ]


def fingerprint(public_key: bytes) -> bytes:
    """Generates a SHA256 hash and formats output to resemble ``ssh-keygen``"""
    h = sha256()
    h.update(public_key)
    return b"SHA256:" + b64encode(h.digest()).rstrip(b"=")


def get_cert_info_object(key_type: PubKeyType) -> OpensshCertificateInfo:
    if key_type == "rsa":
        return OpensshRSACertificateInfo()
    if key_type == "dsa":
        return OpensshDSACertificateInfo()
    if key_type in ("ecdsa-nistp256", "ecdsa-nistp384", "ecdsa-nistp521"):
        return OpensshECDSACertificateInfo()
    if key_type == "ed25519":
        return OpensshED25519CertificateInfo()
    raise ValueError(f"{key_type} is not a valid key type")


def get_option_type(name: str) -> t.Literal["critical", "extension"]:
    if name in _CRITICAL_OPTIONS:
        return "critical"
    if name in _EXTENSIONS:
        return "extension"
    raise ValueError(
        f"{name} is not a valid option. Custom options must start with 'critical:' or 'extension:' to indicate type"
    )


def is_relative_time_string(time_string: str) -> bool:
    return time_string.startswith("+") or time_string.startswith("-")


def parse_option_list(
    option_list: t.Iterable[str],
) -> tuple[list[OpensshCertificateOption], list[OpensshCertificateOption]]:
    critical_options = []
    directives = []
    extensions = []

    for option in option_list:
        if option.lower() in _DIRECTIVES:
            directives.append(option.lower())
        else:
            option_object = OpensshCertificateOption.from_string(option)
            if option_object.type == "critical":
                critical_options.append(option_object)
            else:
                extensions.append(option_object)

    return critical_options, list(set(extensions + apply_directives(directives)))


__all__ = (
    "OpensshCertificate",
    "OpensshCertificateInfo",
    "OpensshCertificateOption",
    "OpensshCertificateTimeParameters",
    "OpensshDSACertificateInfo",
    "OpensshECDSACertificateInfo",
    "OpensshED25519CertificateInfo",
    "OpensshRSACertificateInfo",
    "apply_directives",
    "default_options",
    "fingerprint",
    "get_cert_info_object",
    "get_option_type",
    "is_relative_time_string",
    "parse_option_list",
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_fabric_span_session_source
version_added: "2.11.0"
short_description: Manage Fabric SPAN Sessions Source on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage Switched Port Analyzer (SPAN) Sessions Source on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.4) and later.
author:
- Sabari Jaganathan (@sajagana)
options:
  template:
    description:
    - The name of the template.
    - The template must be a Fabric Monitoring Access Policy template.
    - This parameter or O(template_id) is required.
    type: str
  template_id:
    description:
    - The ID of the template.
    - The template must be a Fabric Monitoring Access Policy template.
    - This parameter or O(template) is required.
    type: str
  span_session_name:
    description:
    - The name of the SPAN Session.
    - This parameter or O(span_session_uuid) is required.
    type: str
  span_session_uuid:
    description:
    - The UUID of the SPAN Session.
    - This parameter or O(span_session_name) is required.
    type: str
  name:
    description:
    - The name of the SPAN Session source.
    type: str
  direction:
    description:
    - The direction of the SPAN Session source.
    - Defaults to O(direction=incoming) when unset during creation.
    type: str
    choices: [ incoming, outgoing, both ]
  span_drop_packets:
    description:
    - The SPAN Drop Packets of the SPAN Session source.
    - SPAN Drop Packets are packets that get dropped when the SPAN destination port cannot handle the volume of mirrored traffic from the source ports.
    - Defaults to O(span_drop_packets=false) when unset during creation.
    - The O(filter_epg) and O(filter_l3out) are not configurable when O(span_drop_packets=true).
    type: bool
  filter_epg:
    description:
    - The Filter EPG of the SPAN Session source.
    - When the Filter EPG is specified in the configuration, the Filter L3Out will be removed.
    - This parameter and O(filter_l3out) are mutually exclusive.
    - Providing an empty dictionary O(filter_epg={}) will remove the filter l3out from the SPAN Session source.
    type: dict
    aliases: [ epg ]
    suboptions:
      uuid:
        description:
        - The UUID of the EPG used to configure the Filter EPG.
        - This parameter or O(filter_epg.reference) is required.
        type: str
        aliases: [ epg_uuid ]
      reference:
        description:
        - The EPG object detail used to configure the Filter EPG.
        - This parameter or O(filter_epg.uuid) is required.
        type: dict
        aliases: [ ref ]
        suboptions:
          name:
            description:
            - The name of the EPG.
            type: str
            required: true
          template:
            description:
            - The name of the template that contains the EPG.
            - This parameter or O(filter_epg.reference.template_id) is required.
            type: str
          template_id:
            description:
            - The ID of the template that contains the EPG.
            - This parameter or O(filter_epg.reference.template) is required.
            type: str
          schema_id:
            description:
            - The ID of the schema that contains the EPG.
            - This parameter or O(filter_epg.reference.schema) is required.
            type: str
          schema:
            description:
            - The name of the schema that contains the EPG.
            - This parameter or O(filter_epg.reference.schema_id) is required.
            type: str
          anp:
            description:
            - The name of the ANP that contains the EPG.
            - This parameter or O(filter_epg.reference.anp_uuid) is required.
            type: str
          anp_uuid:
            description:
            - The UUID of the ANP that contains the EPG.
            - This parameter or O(filter_epg.reference.anp) is required.
            type: str
  filter_l3out:
    description:
    - The Filter L3Out of the SPAN Session source.
    - When the Filter L3Out is specified in the configuration, the Filter EPG will be removed.
    - This parameter and O(filter_epg) are mutually exclusive.
    - The L3Out must be defined in either the L3Out template or directly within the APIC tenant.
    - The Filter L3Out for a SPAN Session source does not support using an L3Out from an application tenant template.
    - Providing an empty dictionary O(filter_l3out={}) will remove the filter l3out from the SPAN Session source.
    type: dict
    aliases: [ l3out ]
    suboptions:
      reference:
        description:
        - The Filter L3Out object detail used to configure the Filter L3Out.
        - This parameter or O(filter_l3out.uuid) is required.
        type: dict
        aliases: [ ref ]
        suboptions:
          name:
            description:
            - The name of the L3Out.
            required: true
            type: str
          tenant:
            description:
            - The name of the tenant. This parameter is used to associate the L3Out from APIC.
            - This parameter or O(filter_l3out.reference.template), or O(filter_l3out.reference.template_id) is required.
            type: str
          template:
            description:
            - The name of the L3Out template.
            - This parameter or O(filter_l3out.reference.template_id), or O(filter_l3out.reference.tenant) is required.
            type: str
          template_id:
            description:
            - The ID of the L3Out template.
            - This parameter or O(filter_l3out.reference.template), or O(filter_l3out.reference.tenant) is required.
            type: str
      uuid:
        description:
        - The UUID of the L3Out.
        - This parameter or O(filter_l3out.reference) is required.
        type: str
        aliases: [ l3out_uuid ]
      vlan_id:
        description:
        - The ID of the VLAN, which is associated with L3Out interface.
        - This parameter is required to configure the Filter L3Out.
        type: int
  access_paths:
    description:
    - The Access Path of the SPAN Session source.
    - Providing a new list of O(access_paths) will completely replace an existing one from the SPAN Session source.
    - Providing an empty list will remove the O(access_paths=[]) from the SPAN Session source.
    type: list
    elements: dict
    suboptions:
      access_path_type:
        description:
        - The type of the Access Path.
        type: str
        choices: [ port, port_channel, virtual_port_channel, vpc_component_pc ]
        aliases: [ type ]
      uuid:
        description:
        - The UUID of the 'Access Port' or 'Port Channel', or 'Virtual Port Channel' which is used to create the Access Path.
        type: str
      node:
        description:
        - The ID of the Node. This parameter is required when O(access_paths.access_path_type=port) or O(access_paths.access_path_type=vpc_component_pc).
        type: int
      interface:
        description:
        - The interface of the Node. This parameter is required to configure the O(access_paths.access_path_type=port).
        type: str
      name:
        description:
        - The name of the 'Port Channel' or 'Virtual Port Channel' which is used to create the Access Path.
        type: str
      template:
        description:
        - The name of the Fabric Resource Policy template.
        - This parameter or O(access_paths.template_id) is required when O(access_paths.access_path_type=port_channel),
          O(access_paths.access_path_type=virtual_port_channel), or O(access_paths.access_path_type=vpc_component_pc).
        type: str
      template_id:
        description:
        - The ID of the Fabric Resource Policy template.
        - This parameter or O(access_paths.template) is required when O(access_paths.access_path_type=port_channel),
          O(access_paths.access_path_type=virtual_port_channel), or O(access_paths.access_path_type=vpc_component_pc).
        type: str
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create the Fabric Monitoring Access Policy template.
- The O(span_session_name) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_fabric_span_session) to create the Fabric SPAN Session.
- The O(filter_epg.reference) must exist before using it with this module in your playbook.
  Use M(cisco.mso.mso_schema_template_anp_epg) to create the EPG.
- The O(filter_l3out.reference) must exist before using it with this module in your playbook.
  Use M(cisco.mso.ndo_l3out_template) to create the L3Out.
- The O(access_paths.name) must exist before using it with this module in your playbook.
  Use M(cisco.mso.ndo_port_channel_interface) to create the Fabric resource port channel interface.
- The O(access_paths.name) must exist before using it with this module in your playbook.
  Use M(cisco.mso.ndo_virtual_port_channel_interface) to create the Fabric resource virtual port channel interface.
seealso:
- module: cisco.mso.ndo_template
- module: cisco.mso.ndo_fabric_span_session
- module: cisco.mso.mso_schema_template_anp_epg
- module: cisco.mso.ndo_port_channel_interface
- module: cisco.mso.ndo_virtual_port_channel_interface
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create the SPAN Session source with access paths
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_2
    direction: outgoing
    span_drop_packets: true
    access_paths:
      - access_path_type: port
        node: 101
        interface: eth1/6
      - access_path_type: port_channel
        name: ansible_test_pc1
        template: ansible_test_fabric_resource
      - access_path_type: virtual_port_channel
        name: ansible_test_vpc1
        template: ansible_test_fabric_resource
      - access_path_type: vpc_component_pc
        name: ansible_test_vpc1
        template: ansible_test_fabric_resource
        node: 101
    state: present
  register: add_ansible_test_source_2

- name: Create the SPAN Session source with access path and filter EPG
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_3
    direction: outgoing
    access_paths:
      - access_path_type: port
        node: 101
        interface: eth1/6
    filter_epg:
      reference:
        schema: ansible_test_schema
        template: template1
        anp: ansible_test_anp
        name: ansible_test_epg1
    state: present

- name: Create the SPAN Session source with access path and filter L3Out
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_4
    direction: outgoing
    access_paths:
      - access_path_type: port
        node: 101
        interface: eth1/1
    filter_l3out:
      reference:
        name: ansible_test_l3out
        template: ansible_test_l3out_template
      vlan_id: 41
    state: present

- name: Create the SPAN Session source with access paths UUID
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template_id: "{{ add_span_session.current.templateId }}"
    span_session_uuid: "{{ add_span_session.current.uuid }}"
    name: ansible_test_source_1
    access_paths:
      - access_path_type: port
        node: 101
        interface: eth1/1
      - access_path_type: port_channel
        uuid: "{{ add_fabric_pc_1.current.uuid }}"
      - access_path_type: virtual_port_channel
        uuid: "{{ add_fabric_vpc1.current.uuid }}"
      - access_path_type: vpc_component_pc
        uuid: "{{ add_fabric_vpc1.current.uuid }}"
        node: 101
    state: present

- name: Update the SPAN Session source Filter EPG using UUID
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_1
    direction: outgoing
    filter_epg:
      uuid: "{{ add_epg.current.epg }}"
    state: present

- name: Update the SPAN Session source Filter L3Out using UUID
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_1
    direction: outgoing
    access_paths:
      - access_path_type: port_channel
        name: ansible_test_pc1
        template: ansible_test_fabric_resource
    filter_l3out:
      uuid: "{{ add_l3out.current.uuid }}"
      vlan_id: 42
    state: present

- name: Query a specific SPAN Session source
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_1
    state: query
  register: query_one

- name: Query all SPAN Session sources
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    state: query
  register: query_all

- name: Delete the SPAN Session source
  cisco.mso.ndo_fabric_span_session_source:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test_mon_pol
    span_session_name: ansible_test_span_session
    name: ansible_test_source_1
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec, epg_object_reference_spec
from ansible_collections.cisco.mso.plugins.module_utils.schemas import MSOSchemas
from ansible_collections.cisco.mso.plugins.module_utils.templates import MSOTemplates
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data, check_if_all_elements_are_none
import copy


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str"),
        template_id=dict(type="str"),
        span_session_name=dict(type="str"),
        span_session_uuid=dict(type="str"),
        name=dict(type="str"),
        direction=dict(type="str", choices=["incoming", "outgoing", "both"]),
        span_drop_packets=dict(type="bool"),
        filter_epg=dict(
            type="dict",
            aliases=["epg"],
            options=dict(
                uuid=dict(type="str", aliases=["epg_uuid"]),
                reference=epg_object_reference_spec(aliases=["ref"]),
            ),
            mutually_exclusive=[("reference", "uuid")],
        ),
        filter_l3out=dict(
            type="dict",
            aliases=["l3out"],
            options=dict(
                reference=dict(
                    type="dict",
                    options=dict(
                        name=dict(type="str", required=True),
                        tenant=dict(type="str"),
                        template=dict(type="str"),
                        template_id=dict(type="str"),
                    ),
                    mutually_exclusive=[("tenant", "template", "template_id")],
                    required_one_of=[["tenant", "template", "template_id"]],
                    aliases=["ref"],
                ),
                uuid=dict(type="str", aliases=["l3out_uuid"]),
                vlan_id=dict(type="int"),
            ),
            mutually_exclusive=[("reference", "uuid")],
            required_by={
                "reference": "vlan_id",
                "uuid": "vlan_id",
            },
        ),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
        access_paths=dict(
            type="list",
            elements="dict",
            options=dict(
                uuid=dict(type="str"),
                node=dict(type="int"),
                interface=dict(type="str"),
                name=dict(type="str"),
                template=dict(type="str"),
                template_id=dict(type="str"),
                access_path_type=dict(type="str", choices=["port", "port_channel", "virtual_port_channel", "vpc_component_pc"], aliases=["type"]),
            ),
            mutually_exclusive=[("uuid", "interface", "name"), ("template", "template_id")],
        ),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[
            ("template", "template_id"),
            ("span_session_name", "span_session_uuid"),
            ("filter_epg", "filter_l3out"),
        ],
        required_if=[
            ["state", "absent", ["name"]],
            ["state", "present", ["name"]],
        ],
        required_one_of=[
            ["template", "template_id"],
            ["span_session_name", "span_session_uuid"],
        ],
    )

    mso = MSOModule(module)
    mso_schemas = MSOSchemas(mso)
    mso_templates = MSOTemplates(mso)

    template_name = module.params.get("template")
    template_id = module.params.get("template_id")
    span_session_name = module.params.get("span_session_name")
    span_session_uuid = module.params.get("span_session_uuid")
    name = module.params.get("name")
    direction = module.params.get("direction")
    span_drop_packets = module.params.get("span_drop_packets")
    filter_epg = module.params.get("filter_epg")
    filter_l3out = module.params.get("filter_l3out")
    access_paths = module.params.get("access_paths")
    state = module.params.get("state")

    if filter_epg is not None and check_if_all_elements_are_none(filter_epg.values()):
        filter_epg = {}

    if filter_l3out is not None and check_if_all_elements_are_none(filter_l3out.values()):
        filter_l3out = {}

    errors = validate_access_paths(access_paths)
    if errors:
        mso.fail_json(msg=", ".join(errors))

    mso_template = MSOTemplate(mso, "monitoring_tenant", template_name, template_id)
    mso_template.validate_template("monitoring")
    site_id = mso_template.template.get("monitoringTemplate").get("sites")[0].get("siteId")

    fabric_span_session = mso_template.get_fabric_span_session(span_session_uuid, span_session_name, fail_module=True)
    match = mso_template.get_fabric_span_session_source(name, fabric_span_session.details.get("sourceGroup", {}).get("sources", []))

    if match and name:
        mso.existing = mso.previous = copy.deepcopy(
            set_fabric_span_session_source_object_details(mso_template, site_id, match.details)
        )  # Query a specific object
    elif match:
        mso.existing = [set_fabric_span_session_source_object_details(mso_template, site_id, obj) for obj in match]  # Query all objects

    if state != "query":
        source_path = "/monitoringTemplate/template/spanSessions/{0}/sourceGroup/sources/{1}".format(fabric_span_session.index, match.index if match else "-")

    ops = []

    if state == "present":
        mso_values = dict(
            name=name,
            direction=direction,
            spanDropPackets=span_drop_packets,
        )

        if filter_epg and (filter_epg.get("reference") or filter_epg.get("uuid")):
            mso_values["epg"] = mso_schemas.get_epg_uuid(filter_epg.get("reference"), filter_epg.get("uuid"))

        if filter_l3out and (filter_l3out.get("reference") or filter_l3out.get("uuid")):
            mso_values["l3out"] = dict(
                encapType="vlan",
                encapValue=filter_l3out.get("vlan_id"),
            )
            if filter_l3out.get("uuid"):
                mso_values["l3out"]["ref"] = filter_l3out.get("uuid")
            elif filter_l3out.get("reference") and filter_l3out.get("reference").get("tenant"):
                mso_values["l3out"]["dn"] = "uni/tn-{0}/out-{1}".format(filter_l3out.get("reference").get("tenant"), filter_l3out.get("reference").get("name"))
            elif filter_l3out.get("reference") and (filter_l3out.get("reference").get("template") or filter_l3out.get("reference").get("template_id")):
                l3out_template = mso_templates.get_template(
                    "l3out", filter_l3out.get("reference").get("template"), filter_l3out.get("reference").get("template_id")
                )
                l3out_match = l3out_template.get_l3out_object(uuid=filter_l3out.get("uuid"), name=filter_l3out.get("reference").get("name"), fail_module=True)
                if l3out_match:
                    mso_values["l3out"]["ref"] = l3out_match.details.get("uuid")

        if access_paths:
            mso_values["accessPaths"] = update_access_paths(mso, site_id, access_paths, mso_templates)

        if match:
            mso_remove_values = []
            proposed_payload = copy.deepcopy(match.details)
            proposed_payload.update({"name": mso_values["name"], "direction": mso_values["direction"], "spanDropPackets": mso_values["spanDropPackets"]})

            if filter_epg == {} and match.details.get("epg"):
                mso_remove_values.append("epg")
                proposed_payload["epg"] = ""
                match.details.pop("epgName")
                match.details.pop("epgTemplateName")
                match.details.pop("epgTemplateId")
                match.details.pop("epgSchemaName")
                match.details.pop("epgSchemaId")
            elif mso_values.get("epg"):
                # When the filter EPG is specified in the configuration, the filter L3Out will be removed.
                # L3Out and EPG cannot be configured simultaneously.
                if match.details.get("l3out"):
                    mso_remove_values.append("l3out")
                proposed_payload["epg"] = mso_values["epg"]

            if filter_l3out == {} and match.details.get("l3out"):
                mso_remove_values.append("l3out")
                proposed_payload["l3out"] = {}
                match.details["l3out"] = {}
            elif mso_values.get("l3out"):
                # When the filter L3Out is specified in the configuration, the filter EPG will be removed.
                # L3Out and EPG cannot be configured simultaneously.
                if match.details.get("epg"):
                    mso_remove_values.append("epg")
                proposed_payload["l3out"] = mso_values["l3out"]

            if access_paths == [] and match.details.get("accessPaths"):
                mso_remove_values.append("accessPaths")
                proposed_payload["accessPaths"] = []
                match.details["accessPaths"] = []
            elif mso_values.get("accessPaths"):
                proposed_payload["accessPaths"] = mso_values["accessPaths"]

            mso.sanitize(proposed_payload, collate=True)
            append_update_ops_data(ops, match.details, source_path, mso_values, mso_remove_values)
        else:
            mso.sanitize(mso_values)
            ops.append(dict(op="add", path=source_path, value=mso.sent))

    elif state == "absent" and match:
        ops.append(dict(op="remove", path=source_path))

    if not module.check_mode and ops:
        mso_template.template = mso.request(mso_template.template_path, method="PATCH", data=ops)
        fabric_span_session = mso_template.get_fabric_span_session(span_session_uuid, span_session_name, fail_module=True)
        match = mso_template.get_fabric_span_session_source(name, fabric_span_session.details.get("sourceGroup", {}).get("sources", []))
        if match:
            mso.existing = set_fabric_span_session_source_object_details(mso_template, site_id, match.details)  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.existing = set_fabric_span_session_source_object_details(mso_template, site_id, mso.proposed) if state == "present" else {}

    mso.exit_json()


def validate_access_paths(access_paths):
    if access_paths:
        errors = []
        for index, path in enumerate(access_paths):
            access_path_type = path.get("access_path_type")
            uuid = path.get("uuid")
            node = path.get("node")
            interface = path.get("interface")
            name = path.get("name")
            template = path.get("template")
            template_id = path.get("template_id")

            # Validate based on access_path_type
            if access_path_type == "port":
                if not uuid and not (node and interface):
                    errors.append(
                        "Access path {0}: when the access_path_type='port', either 'uuid' or both 'node' and 'interface' must be provided.".format(index + 1)
                    )

            elif access_path_type == "port_channel":
                if not uuid and not (name and template) and not (name and template_id):
                    errors.append(
                        (
                            "Access path {0}: when the access_path_type='port_channel', either 'uuid',"
                            + " or both 'name' and 'template', or both 'name' and 'template_id' must be provided."
                        ).format(index + 1)
                    )

            elif access_path_type == "virtual_port_channel":
                if not uuid and not (name and template) and not (name and template_id):
                    errors.append(
                        (
                            "Access path {0}: when the access_path_type 'virtual_port_channel', either 'uuid',"
                            + " or both 'name' and 'template', or both 'name' and 'template_id' must be provided."
                        ).format(index + 1)
                    )

            elif access_path_type == "vpc_component_pc":
                if not (uuid and node) and not (name and template and node) and not (name and template_id and node):
                    errors.append(
                        (
                            "Access path {0}: when the access_path_type 'vpc_component_pc', either both 'uuid' and 'node',"
                            + " or all of 'name', 'template', and 'node',"
                            + " or all of 'name', 'template_id', and 'node' must be provided."
                        ).format(index + 1)
                    )
        return errors


def get_access_path_payload(mso_templates, access_path_config, access_path_type, resource_type):
    uuid = access_path_config.get("uuid")
    if not uuid:
        fabric_template = mso_templates.get_template("fabric_resource", access_path_config.get("template"), access_path_config.get("template_id"))
        uuid = fabric_template.get_template_policy_uuid("fabric_resource", access_path_config.get("name"), resource_type)

    return (
        dict(vpcComponentPc=[dict(vpc=uuid, node=str(access_path_config.get("node")))]) if access_path_type == "vpc_component_pc" else {resource_type: [uuid]}
    )


def update_access_paths(mso, site_id, access_paths, mso_templates):
    updated_paths = []
    for access_path in access_paths:
        if access_path.get("access_path_type") == "port":
            port_uuid = access_path.get("uuid")
            if not port_uuid:
                port_uuid = mso.get_site_interface_details(
                    site_id=site_id,
                    uuid=None,
                    node=access_path.get("node"),
                    port=access_path.get("interface"),
                ).get("uuid")
            updated_paths.append(dict(accessInterfaces=[port_uuid]))

        elif access_path.get("access_path_type") == "port_channel":
            updated_paths.append(get_access_path_payload(mso_templates, access_path, "port_channel", "portChannels"))

        elif access_path.get("access_path_type") == "virtual_port_channel":
            updated_paths.append(get_access_path_payload(mso_templates, access_path, "virtual_port_channel", "virtualPortChannels"))

        elif access_path.get("access_path_type") == "vpc_component_pc":
            updated_paths.append(get_access_path_payload(mso_templates, access_path, "vpc_component_pc", "virtualPortChannels"))

    if len(set([str(path) for path in updated_paths])) == len(updated_paths):
        return updated_paths
    else:
        mso.fail_json(msg="Remove duplicate entries from the access_paths: {0}".format(access_paths))


def set_fabric_span_session_source_object_details(mso_template, site_id, source):
    if source:
        for access_path in source.get("accessPaths", []):  # Adding the object reference name to use the update_config_with_template_and_references function
            if access_path.get("accessInterfaces") and isinstance(access_path.get("accessInterfaces")[0], str):
                access_path.get("accessInterfaces")[0] = mso_template.mso.get_site_interface_details(site_id, access_path.get("accessInterfaces")[0])
            elif access_path.get("portChannels") and isinstance(access_path.get("portChannels")[0], str):
                access_path.get("portChannels")[0] = dict(portChannel=access_path.get("portChannels")[0])
            elif access_path.get("virtualPortChannels") and isinstance(access_path.get("virtualPortChannels")[0], str):
                access_path.get("virtualPortChannels")[0] = dict(virtualPortChannel=access_path.get("virtualPortChannels")[0])

        reference_details = {
            "filterEPG": {
                "name": "epgName",
                "reference": "epg",
                "type": "epg",
                "template": "epgTemplateName",
                "templateId": "epgTemplateId",
                "schema": "epgSchemaName",
                "schemaId": "epgSchemaId",
            },
            "portChannels": {
                "name": "portChannelName",
                "reference": "portChannel",
                "type": "portChannel",
                "template": "portChannelTemplateName",
                "templateId": "portChannelTemplateId",
            },
            "virtualPortChannels": {
                "name": "virtualPortChannelName",
                "reference": "virtualPortChannel",
                "type": "virtualPortChannel",
                "template": "virtualPortChannelTemplateName",
                "templateId": "virtualPortChannelTemplateId",
            },
            "vpcComponentPc": {
                "name": "vpcComponentPcName",
                "reference": "vpc",
                "type": "virtualPortChannel",
                "template": "vpcComponentPcTemplateName",
                "templateId": "vpcComponentPcTemplateId",
            },
            "l3out": {
                "name": "l3outName",
                "reference": "ref",
                "type": "l3out",
                "template": "l3outTemplateName",
                "templateId": "l3outTemplateId",
            },
        }

        mso_template.update_config_with_template_and_references(source, reference_details, True)
    return source


if __name__ == "__main__":
    main()



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025 Felix Fontein <felix@fontein.de>

# /// script
# dependencies = ["nox>=2025.02.09", "antsibull-nox"]
# ///

import sys

import nox

try:
    import antsibull_nox
    from antsibull_nox.cli import run as run_antsibull_nox
except ImportError:
    print("You need to install antsibull-nox in the same Python environment as nox.")
    sys.exit(1)


antsibull_nox.load_antsibull_nox_toml()


@nox.session(name="create-certificates", default=False)
def create_certificates(session: nox.Session) -> None:
    """
    Regenerate some vendored certificates.
    """
    session.install("cryptography<39.0.0")  # we want support for SHA1 signatures
    session.run("python", "tests/create-certificates.py")
    session.warn(
        "Note that you need to modify some values in tests/integration/targets/x509_certificate_info/tasks/impl.yml"
        " and tests/integration/targets/filter_x509_certificate_info/tasks/impl.yml!"
    )


@nox.session(name="update-azp-config", python=False)
def update_azp_config(session: nox.Session) -> None:
    command = [
        "antsibull-nox",
        "update-azp-config",
        "--min-ansible-core",
        "2.19",
    ]
    if antsibull_nox.IN_CI:
        command.append("--fail-on-change")
    session.debug(" ".join(command))
    result = run_antsibull_nox(command)
    if result != 0:
        session.error(f"Execution failed with status code {result}")


# Allow to run the noxfile with `python noxfile.py`, `pipx run noxfile.py`, or similar.
# Requires nox >= 2025.02.09
if __name__ == "__main__":
    nox.main()



#!/usr/bin/python
# Copyright (c) 2018, Juan Manuel Parrilla <jparrill@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipa_vault
author: Juan Manuel Parrilla (@jparrill)
short_description: Manage FreeIPA vaults
description:
  - Add, modify and delete vaults and secret vaults.
  - KRA service should be enabled to use this module.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  cn:
    description:
      - Vault name.
      - Can not be changed as it is the unique identifier.
    required: true
    aliases: ["name"]
    type: str
  description:
    description:
      - Description.
    type: str
  ipavaulttype:
    description:
      - Vault types are based on security level.
    default: "symmetric"
    choices: ["asymmetric", "standard", "symmetric"]
    aliases: ["vault_type"]
    type: str
  ipavaultpublickey:
    description:
      - Public key.
    aliases: ["vault_public_key"]
    type: str
  ipavaultsalt:
    description:
      - Vault Salt.
    aliases: ["vault_salt"]
    type: str
  username:
    description:
      - Any user can own one or more user vaults.
      - Mutually exclusive with O(service).
    aliases: ["user"]
    type: list
    elements: str
  service:
    description:
      - Any service can own one or more service vaults.
      - Mutually exclusive with O(user).
    type: str
  state:
    description:
      - State to ensure.
    default: "present"
    choices: ["absent", "present"]
    type: str
  replace:
    description:
      - Force replace the existent vault on IPA server.
    type: bool
    default: false
    choices: ["True", "False"]
  validate_certs:
    description:
      - Validate IPA server certificates.
    type: bool
    default: true
extends_documentation_fragment:
  - community.general._ipa.documentation
  - community.general._ipa.connection_notes
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Ensure vault is present
  community.general.ipa_vault:
    name: vault01
    vault_type: standard
    user: user01
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure vault is present for Admin user
  community.general.ipa_vault:
    name: vault01
    vault_type: standard
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure vault is absent
  community.general.ipa_vault:
    name: vault01
    vault_type: standard
    user: user01
    state: absent
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Modify vault if already exists
  community.general.ipa_vault:
    name: vault01
    vault_type: standard
    description: "Vault for test"
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
    replace: true

- name: Get vault info if already exists
  community.general.ipa_vault:
    name: vault01
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
"""

RETURN = r"""
vault:
  description: Vault as returned by IPA API.
  returned: always
  type: dict
"""

import traceback

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ipa import IPAClient, ipa_argument_spec


class VaultIPAClient(IPAClient):
    def __init__(self, module, host, port, protocol):
        super().__init__(module, host, port, protocol)

    def vault_find(self, name):
        return self._post_json(method="vault_find", name=None, item={"all": True, "cn": name})

    def vault_add_internal(self, name, item):
        return self._post_json(method="vault_add_internal", name=name, item=item)

    def vault_mod_internal(self, name, item):
        return self._post_json(method="vault_mod_internal", name=name, item=item)

    def vault_del(self, name):
        return self._post_json(method="vault_del", name=name)


def get_vault_dict(description=None, vault_type=None, vault_salt=None, vault_public_key=None, service=None):
    vault = {}

    if description is not None:
        vault["description"] = description
    if vault_type is not None:
        vault["ipavaulttype"] = vault_type
    if vault_salt is not None:
        vault["ipavaultsalt"] = vault_salt
    if vault_public_key is not None:
        vault["ipavaultpublickey"] = vault_public_key
    if service is not None:
        vault["service"] = service
    return vault


def get_vault_diff(client, ipa_vault, module_vault, module):
    return client.get_diff(ipa_data=ipa_vault, module_data=module_vault)


def ensure(module, client):
    state = module.params["state"]
    name = module.params["cn"]
    # user = module.params["username"]  TODO is this really not needed?
    replace = module.params["replace"]

    module_vault = get_vault_dict(
        description=module.params["description"],
        vault_type=module.params["ipavaulttype"],
        vault_salt=module.params["ipavaultsalt"],
        vault_public_key=module.params["ipavaultpublickey"],
        service=module.params["service"],
    )
    ipa_vault = client.vault_find(name=name)

    changed = False
    if state == "present":
        if not ipa_vault:
            # New vault
            changed = True
            if not module.check_mode:
                ipa_vault = client.vault_add_internal(name, item=module_vault)
        else:
            # Already exists
            if replace:
                diff = get_vault_diff(client, ipa_vault, module_vault, module)
                if len(diff) > 0:
                    changed = True
                    if not module.check_mode:
                        data = {}
                        for key in diff:
                            data[key] = module_vault.get(key)
                        client.vault_mod_internal(name=name, item=data)

    else:
        if ipa_vault:
            changed = True
            if not module.check_mode:
                client.vault_del(name)

    return changed, client.vault_find(name=name)


def main():
    argument_spec = ipa_argument_spec()
    argument_spec.update(
        cn=dict(type="str", required=True, aliases=["name"]),
        description=dict(type="str"),
        ipavaulttype=dict(
            type="str", default="symmetric", choices=["standard", "symmetric", "asymmetric"], aliases=["vault_type"]
        ),
        ipavaultsalt=dict(type="str", aliases=["vault_salt"]),
        ipavaultpublickey=dict(type="str", aliases=["vault_public_key"]),
        service=dict(type="str"),
        replace=dict(type="bool", default=False, choices=[True, False]),
        state=dict(type="str", default="present", choices=["present", "absent"]),
        username=dict(type="list", elements="str", aliases=["user"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec, supports_check_mode=True, mutually_exclusive=[["username", "service"]]
    )

    client = VaultIPAClient(
        module=module,
        host=module.params["ipa_host"],
        port=module.params["ipa_port"],
        protocol=module.params["ipa_prot"],
    )
    try:
        client.login(username=module.params["ipa_user"], password=module.params["ipa_pass"])
        changed, vault = ensure(module, client)
        module.exit_json(changed=changed, vault=vault)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Ospf_interfaces parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _tmplt_ospf_int_authentication(config_data):
    if "authentication_v2" in config_data:
        command = "ip ospf authentication"
        if "message_digest" in config_data["authentication_v2"]:
            command += " message-digest"
        return command
    if "authentication_v3" in config_data:
        command = "ospfv3 authentication ipsec spi "
        command += "{spi} {algorithm}".format(**config_data["authentication_v3"])
        if "passphrase" in config_data["authentication_v3"]:
            command += " passphrase"
        if "keytype" in config_data["authentication_v3"]:
            command += " {keytype}".format(**config_data["authentication_v3"])
        if "passphrase" not in config_data["authentication_v3"]:
            command += " {key}".format(**config_data["authentication_v3"])
        else:
            command += " {passphrase}".format(**config_data["authentication_v3"])
        return command


def _tmplt_ospf_int_encryption_v3(config_data):
    if "encryption" in config_data:
        command = "ospfv3 encryption ipsec spi ".format(**config_data)
        command += "{spi} esp {encryption} {algorithm}".format(**config_data["encryption"])
        if "passphrase" in config_data["encryption"]:
            command += " passphrase"
        if "keytype" in config_data["encryption"]:
            command += " {keytype}".format(**config_data["encryption"])
        if "passphrase" not in config_data["encryption"]:
            command += " {key}".format(**config_data["encryption"])
        else:
            command += " {passphrase}".format(**config_data["encryption"])
        return command


def _tmplt_ospf_int_authentication_key(config_data):
    if "authentication_key" in config_data:
        command = "ip ospf authentication-key"
        if "encryption" in config_data["authentication_key"]:
            command += " {encryption} {key}".format(**config_data["authentication_key"])
        else:
            command += " {key}".format(**config_data["authentication_key"])
        return command


def _tmplt_ospf_int_cost(config_data):
    if config_data["afi"] == "ipv4":
        command = "ip ospf cost {cost}".format(**config_data)
    else:
        command = "ospfv3 cost {cost}".format(**config_data)
    return command


def _tmplt_ospf_int_bfd(config_data):
    if config_data["afi"] == "ipv4":
        command = "ip ospf bfd"
    else:
        command = "ospfv3 bfd"
    return command


def _tmplt_ospf_int_hello_interval(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} hello-interval {hello_interval}".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf hello-interval {hello_interval}".format(**config_data)
        else:
            command = "ospfv3 hello-interval {hello_interval}".format(**config_data)
    return command


def _tmplt_ospf_int_mtu_ignore(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} mtu-ignore".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf mtu-ignore"
        else:
            command = "ospfv3 mtu-ignore"
    return command


def _tmplt_ospf_int_network(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} network {network}".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf network {network}".format(**config_data)
        else:
            command = "ospfv3 network {network}".format(**config_data)
    return command


def _tmplt_ospf_int_priority(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} priority {priority}".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf priority {priority}".format(**config_data)
        else:
            command = "ospfv3 priority {priority}".format(**config_data)
    return command


def _tmplt_ospf_int_retransmit_interval(config_data):
    if "ip_params" in config_data:
        command = (
            "ospfv3 "
            + config_data["ip_params"]["afi"]
            + " retransmit-interval "
            + str(config_data["ip_params"]["retransmit_interval"])
        )
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf retransmit-interval {retransmit_interval}".format(**config_data)
        else:
            command = "ospfv3 retransmit-interval {retransmit_interval}".format(**config_data)
    return command


def _tmplt_ospf_int_transmit_delay(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} transmit-delay {transmit_delay}".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf transmit-delay {transmit_delay}".format(**config_data)
        else:
            command = "ospfv3 transmit-delay {transmit_delay}".format(**config_data)
    return command


def _tmplt_ospf_int_dead_interval(config_data):
    if "ip_params" in config_data:
        command = "ospfv3 {afi} dead-interval {dead_interval}".format(**config_data["ip_params"])
    else:
        if config_data["afi"] == "ipv4":
            command = "ip ospf dead-interval {dead_interval}".format(**config_data)
        else:
            command = "ospfv3 dead-interval {dead_interval}".format(**config_data)
    return command


class Ospf_interfacesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Ospf_interfacesTemplate, self).__init__(
            lines=lines,
            tmplt=self,
            module=module,
        )

    PARSERS = [
        {
            "name": "interfaces",
            "getval": re.compile(
                r"""
                ^interface
                \s+(?P<name>\S+)
                $""",
                re.VERBOSE,
            ),
            "setval": "interface {{ name }}",
            "result": {"name": "{{ name }}"},
            "shared": True,
        },
        {
            "name": "area",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+area
                \s+(?P<area_id>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": "ip ospf area {{ area.area_id }}",
            "compval": "area",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "area": {"area_id": "{{ area_id }}"},
                    },
                },
            },
        },
        {
            "name": "authentication_v2",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+authentication
                \s*(?P<message_digest>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_authentication,
            "compval": "authentication_v2",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "authentication_v2": {
                            "set": "{{ True if message_digest is undefined }}",
                            "message_digest": "{{ True if message_digest is defined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "authentication_v3",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+authentication
                \s+ipsec
                \s+spi
                \s+(?P<val>\d+)
                \s+(?P<algorithm>md5|sha1)
                \s*(?P<passphrase>passphrase)*
                \s*(?P<type>0|7)*
                \s*(?P<line>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_authentication,
            "compval": "authentication_v3",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "authentication_v3": {
                            "spi": "{{ val }}",
                            "algorithm": "{{ algorithm }}",
                            "keytype": "{{ type }}",
                            "passphrase": "{{ line if passphrase is defined }}",
                            "key": "{{ str(line) if passphrase is undefined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "authentication_key",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+authentication-key
                \s*(?P<encryption>\d+)*
                \s*(?P<line>\S+)*
                $""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_authentication_key,
            "compval": "authentication_key",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "authentication_key": {
                            "encryption": "{{ encryption }}",
                            "key": "{{ line }}",
                        },
                    },
                },
            },
        },
        {
            "name": "bfd",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+bfd
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_bfd,
            "compval": "bfd",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "bfd": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "deadinterval",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+dead-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_dead_interval,
            "compval": "dead_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "dead_interval": "{{ interval }}",
                    },
                },
            },
        },
        {
            "name": "encryption",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+encryption
                \s+ipsec
                \s+spi
                \s+(?P<val>\d+)
                \s+esp
                \s+(?P<encryption>\S+)
                \s*(?P<algorithm>md5|sha1)
                \s*(?P<passphrase>passphrase)
                \s*(?P<type>0|7)
                \s*(?P<line>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_encryption_v3,
            "compval": "encryption_v3",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "encryption_v3": {
                            "spi": "{{ val }}",
                            "encryption": "{{ encryption }}",
                            "algorithm": "{{ algorithm }}",
                            "keytype": "{{ type }}",
                            "passphrase": "{{ line if passphrase is defined }}",
                            "key": "{{ str(line) if passphrase is undefined }}",
                        },
                    },
                },
            },
        },
        {
            "name": "hellointerval",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+hello-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_hello_interval,
            "compval": "hello_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "hello_interval": "{{ interval }}",
                    },
                },
            },
        },
        {
            "name": "bfd",
            "getval": re.compile(
                r"""
                \s+ospfv3
                \s+bfd
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_bfd,
            "compval": "bfd",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "bfd": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "cost",
            "getval": re.compile(
                r"""
                \s+ip
                \s+ospf
                \s+cost
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_cost,
            "compval": "cost",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "cost": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "cost",
            "getval": re.compile(
                r"""
                \s+ospfv3
                \s+cost
                \s+(?P<cost>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_cost,
            "compval": "cost",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "cost": "{{ cost }}",
                    },
                },
            },
        },
        {
            "name": "deadinterval",
            "getval": re.compile(
                r"""
                \s+ospfv3
                \s+dead-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_dead_interval,
            "compval": "dead_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "dead_interval": "{{ interval }}",
                    },
                },
            },
        },
        {
            "name": "hellointerval",
            "getval": re.compile(
                r"""
                \s+ospfv3
                \s+hello-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_hello_interval,
            "compval": "hello_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "hello_interval": "{{ interval }}",
                    },
                },
            },
        },
        {
            "name": "ip_params_area",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+area
                \s+(?P<area_id>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 {{ ip_params.afi }} area {{ ip_params.area.area_id }}",
            "compval": "ip_params.area",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "area": {"area_id": "{{ area_id }}"},
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_bfd",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+bfd
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 {{ afi }} bfd",
            "compval": "ip_params.bfd",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "bfd": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_cost",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+cost
                \s+(?P<cost>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 {{ afi }} cost {{ cost }}",
            "compval": "ip_params.cost",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "cost": "{{ cost }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_dead_interval",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+dead-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_dead_interval,
            "compval": "ip_params.dead_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "dead_interval": "{{ interval }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_hello_interval",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+hello-interval
                \s+(?P<interval>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 {{ ip_params.afi }} hello-interval {{ ip_params.hello_interval }}",
            "compval": "ip_params.hello_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "hello_interval": "{{ interval }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_mtu_ignore",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+(?P<mtu>mtu-ignore)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_mtu_ignore,
            "compval": "ip_params.mtu_ignore",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "mtu_ignore": "{{ True if mtu is defined}}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_network",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+network
                \s+(?P<val>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_network,
            "compval": "ip_params.network",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "network": "{{ val }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_priority",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+priority
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_priority,
            "compval": "ip_params.priority",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "priority": "{{ val }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_passive_interface",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+passive-interface
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 {{ ip_params.afi }} passive-interface",
            "compval": "ip_params.passive_interface",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "passive_interface": "{{ True }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_retransmit_interval",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+retransmit-interval
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_retransmit_interval,
            "compval": "ip_params.retransmit_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "retransmit_interval": "{{ val }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "ip_params_transmit_delay",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+(?P<afi>ipv4|ipv6)
                \s+transmit-delay
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_transmit_delay,
            "compval": "ip_params.transmit_delay",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "ip_params": {
                            "{{ afi }}": {
                                "afi": "{{ afi }}",
                                "transmit_delay": "{{ val }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "mtu_ignore",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+mtu-ignore
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_mtu_ignore,
            "compval": "mtu_ignore",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "mtu_ignore": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "network",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+network
                \s+(?P<interface>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_network,
            "compval": "network",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "network": "{{ interface }}",
                    },
                },
            },
        },
        {
            "name": "priority",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+priority
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_priority,
            "compval": "priority",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "priority": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "passive_interface",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+passive-interface
                *$""",
                re.VERBOSE,
            ),
            "setval": "ospfv3 passive-interface",
            "compval": "passive_interface",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "passive_interface": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "retransmit_interval",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+retransmit-interval
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_retransmit_interval,
            "compval": "retransmit_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "retransmit_interval": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "transmit_delay",
            "getval": re.compile(
                r"""
                \s*ospfv3
                \s+transmit-delay
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_transmit_delay,
            "compval": "transmit_delay",
            "result": {
                "address_family": {
                    "{{ 'ipv6' }}": {
                        "afi": '{{ "ipv6" }}',
                        "transmit_delay": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "mtu_ignore",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+mtu-ignore
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_mtu_ignore,
            "compval": "mtu_ignore",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "mtu_ignore": "{{ True }}",
                    },
                },
            },
        },
        {
            "name": "network",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+network
                \s+(?P<interface>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_network,
            "compval": "network",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "network": "{{ interface }}",
                    },
                },
            },
        },
        {
            "name": "priority",
            "getval": re.compile(
                r"""
                \s*ip ospf
                \s+priority
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_priority,
            "compval": "priority",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "priority": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "retransmit_interval",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+retransmit-interval
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_retransmit_interval,
            "compval": "retransmit_interval",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "retransmit_interval": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "transmit_delay",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+transmit-delay
                \s+(?P<val>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": _tmplt_ospf_int_transmit_delay,
            "compval": "transmit_delay",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "transmit_delay": "{{ val }}",
                    },
                },
            },
        },
        {
            "name": "message_digest_key",
            "getval": re.compile(
                r"""
                \s*ip
                \s+ospf
                \s+message-digest-key
                \s+(?P<id>\d+)
                \s+md5
                \s*(?P<type>0|7)*
                \s+(?P<line>\S+)
                *$""",
                re.VERBOSE,
            ),
            "setval": "ip ospf message-digest-key {{ message_digest_key.key_id }} md5 {{ message_digest_key.encryption }} {{ message_digest_key.key }}",
            "compval": "message_digest_key",
            "result": {
                "address_family": {
                    "{{ 'ipv4' }}": {
                        "afi": '{{ "ipv4" }}',
                        "message_digest_key": {
                            "key_id": "{{ id }}",
                            "encryption": "{{ type }}",
                            "key": "{{ line }}",
                        },
                    },
                },
            },
        },
    ]



# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # Standard files documentation fragment
    DOCUMENTATION = r"""
options:
  api_url:
    description:
      - The resolvable endpoint for the API.
    type: str
  api_username:
    description:
      - The username to use for authentication against the API.
    type: str
  api_password:
    description:
      - The password to use for authentication against the API.
    type: str
  validate_certs:
    description:
      - Whether or not to validate SSL certs when supplying a HTTPS endpoint.
    type: bool
    default: true
"""
